# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import logging
import math
import re
import textwrap
import warnings
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from itertools import zip_longest
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Collection,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import attr
import numpy as np
import pandas as pd
import pint

from qslib import scpi_commands
from qslib.data import FilterSet

from ._util import *
from .base import RunStatus
from .scpi_commands import SCPICommand, SCPICommandLike
from .version import __version__

if TYPE_CHECKING:  # pragma: no cover
    import matplotlib.pyplot as plt

NZONES = 6

UR: pint.UnitRegistry = pint.UnitRegistry(
    autoconvert_offset_to_baseunit=True, auto_reduce_dimensions=True
)


log = logging.getLogger(__name__)


def _wrap_seconds(val: int | float | str | pint.Quantity[Any]) -> pint.Quantity[Any]:
    if isinstance(val, str):
        uv = UR(val)
        uv.check("seconds")
    elif isinstance(val, pint.Quantity):
        uv = val
        uv.check("seconds")
    else:
        uv = UR.Quantity(val, "seconds")
    return uv


def _wrap_degC(val: int | float | str | pint.Quantity[Any]) -> pint.Quantity[Any]:
    if isinstance(val, str):
        uv = UR(val)
        uv.check("degC")
    elif isinstance(val, pint.Quantity):
        uv = val
        uv.check("degC")
    else:
        uv = UR.Quantity(val, "degC")
    return uv


def _wrap_delta_degC(val: int | float | str | pint.Quantity[Any]) -> pint.Quantity[Any]:
    if isinstance(val, str):
        uv = UR(val)
        uv.check("delta_degC")
    elif isinstance(val, pint.Quantity):
        uv = val
        uv.check("delta_degC")
    else:
        uv = UR.Quantity(val, "delta_degC")
    return uv


_ZEROTEMPDELTA = UR.Quantity(0.0, "delta_degC")


def _wrap_delta_degC_or_zero(
    val: int | float | str | pint.Quantity[Any] | None,
) -> pint.Quantity[Any]:
    if val is None:
        return _ZEROTEMPDELTA
    else:
        return _wrap_delta_degC(val)


def _wrap_degC_or_none(
    val: int | float | str | pint.Quantity[Any] | None,
) -> pint.Quantity[Any] | None:
    if val is None:
        return None
    else:
        return _wrap_degC(val)


def _wrapunitmaybelist_degC(
    val: int
    | float
    | str
    | pint.Quantity[Any]
    | Sequence[int | float | str | pint.Quantity[Any]],
) -> pint.Quantity[Any]:
    unit: pint.Unit = cast(pint.Unit, UR("degC"))

    if isinstance(val, pint.Quantity):
        uv = val
        uv.check(unit)
    elif isinstance(val, str):
        uv = UR(val)
        uv.check(unit)
    elif isinstance(val, Sequence):
        m = []
        for x in val:
            if isinstance(x, pint.Quantity):
                m.append(x.to(unit).magnitude)
            elif isinstance(x, str):
                m.append(UR(x).to(unit).magnitude)
            else:
                m.append(x)
        return UR.Quantity(m, unit)
    else:
        uv = UR.Quantity(val, unit)
    return uv


def _durformat(time: pint.Quantity[int]) -> str:
    """Convert time in seconds to a nice string"""
    time_s: int = time.to(UR.seconds).magnitude
    s = ""
    # <= 2 minutes: stay as seconds
    if time_s <= 2 * 60:
        s = "{}s".format(int(time_s)) + s
        return s

    if (time_s % 60) != 0:
        s = "{}s".format(int(time_s % 60)) + s
    rtime = int(time_s // 60)
    # <= 2 hours: stay as minutes
    if time_s <= 2 * 60 * 60:
        s = "{}m".format(rtime) + s
        return s

    if (rtime % 60) != 0:
        s = "{}m".format(rtime % 60) + s
    rtime = rtime // 60
    # <= 3 days: stay as hours
    if time_s <= 3 * 24 * 3600:
        if (rtime % 24) != 0:
            s = "{}h".format(rtime) + s
        return s

    s = "{}h".format(rtime % 24) + s
    rtime = rtime // 24
    # days without bound
    s = "{}d".format(rtime) + s
    return s


T = TypeVar("T")


class ProtoCommand(ABC):
    @abstractmethod
    def to_scpicommand(self, **kwargs: Any) -> SCPICommand:  # pragma: no cover
        ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        ...

    @classmethod
    def from_scpicommand(cls: Type[T], sc: SCPICommand) -> T:  # pragma: no cover
        ...

    _names: ClassVar[Sequence[str]] = tuple()


@attr.define
class Ramp(ProtoCommand):
    """Ramps temperature to a new setting."""

    temperature: pint.Quantity[np.ndarray] = attr.field(
        converter=_wrapunitmaybelist_degC, on_setattr=attr.setters.convert
    )
    increment: pint.Quantity[float] = attr.field(
        converter=_wrap_delta_degC_or_zero,
        on_setattr=attr.setters.convert,
        default=_ZEROTEMPDELTA,
    )
    incrementcycle: int = 1
    incrementstep: int = 1
    rate: float = 100.0  # This is a percent
    cover: pint.Quantity[float] | None = attr.field(
        converter=_wrap_degC_or_none,
        on_setattr=attr.setters.convert,
        default=None,
    )
    _names: ClassVar[Sequence[str]] = ("RAMP",)

    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
        opts = {}

        if self.increment != _ZEROTEMPDELTA:
            opts["increment"] = self.increment.to("delta_degC").magnitude
        if self.incrementcycle != 1:
            opts["incrementcycle"] = self.incrementcycle
        if self.incrementstep != 1:
            opts["incrementstep"] = self.incrementstep
        if self.rate != 100.0:
            opts["rate"] = self.rate
        if self.cover is not None:
            opts["cover"] = self.cover.to("degC").magnitude

        return SCPICommand(
            "RAMP", *self.temperature.to("degC").magnitude, comment=None, **opts
        )

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> Ramp:
        return Ramp(UR.Quantity(sc.args, "degC"), **sc.opts)  # type: ignore


@dataclass
class Exposure(ProtoCommand):
    """Modifies exposure settings."""

    # We don't support persistent... it doesn't seem safe
    settings: Sequence[Tuple[FilterSet, Sequence[int]]]
    state: str = "HoldAndCollect"
    _names: ClassVar[Sequence[str]] = ("EXP", "EXPOSURE")

    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
        settingstrings = [
            k.hacform + "," + ",".join(str(x) for x in v) for k, v in self.settings
        ]

        return SCPICommand(
            "EXP",
            *[[k.hacform] + [str(x) for x in v] for k, v in self.settings],
            state=self.state,
        )

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> Exposure:
        filts = [
            (FilterSet.fromstring(f"{x[1]},{x[0]},{x[2]}"), [int(y) for y in x[3:]])
            for x in cast(Sequence[str], sc.args)
        ]
        return Exposure(filts, **sc.opts)  # type: ignore


def _filtersequence(x: Sequence[str | FilterSet]) -> Sequence[FilterSet]:
    return [FilterSet.fromstring(f) for f in x]


@attr.define
class HACFILT(ProtoCommand):
    """Sets filters for :class:`HoldAndCollect` ."""

    filters: Sequence[FilterSet] = attr.field(
        converter=_filtersequence,
        on_setattr=attr.setters.convert,
    )
    _default_filters: Sequence[FilterSet] = attr.field(factory=lambda: [])
    _names: ClassVar[Sequence[str]] = ("HoldAndCollectFILTer", "HACFILT")

    def to_scpicommand(
        self, default_filters: Sequence[FilterSet] | None = None, **kwargs: None
    ) -> SCPICommand:
        if default_filters is None:
            default_filters = []
        if not default_filters and not self.filters:
            raise ValueError("Protocol must have default filters set.")
        if not self.filters:
            comment = "qslib:default_filters"
        else:
            comment = None
        return SCPICommand(
            "HACFILT",
            *(f.hacform for f in (self.filters if self.filters else default_filters)),
            comment=comment,
        )

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> HACFILT:
        c = HACFILT([FilterSet.fromstring(x) for x in cast(Iterable[str], sc.args)])
        if sc.comment and "qslib:default_filters" in sc.comment:
            c._default_filters = c.filters
            c.filters = []
        return c


@dataclass
class HoldAndCollect(ProtoCommand):
    """A protocol hold (for a time) and collect (set by HACFILT) command."""

    time: pint.Quantity[int]
    increment: pint.Quantity[int] = UR.Quantity(0, "seconds")
    incrementcycle: int = 1
    incrementstep: int = 1
    tiff: bool = False
    quant: bool = True
    pcr: bool = False
    _names: ClassVar[Sequence[str]] = ("HoldAndCollect",)

    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
        opts = {}
        if self.increment != HoldAndCollect.increment:
            opts["increment"] = self.increment.to("seconds").magnitude
        if self.incrementcycle != HoldAndCollect.incrementcycle:
            opts["incrementcycle"] = self.incrementcycle
        if self.incrementstep != HoldAndCollect.incrementstep:
            opts["incrementstep"] = self.incrementstep
        opts["tiff"] = self.tiff
        opts["quant"] = self.quant
        opts["pcr"] = self.pcr
        return SCPICommand(
            "HoldAndCollect",
            int(self.time.to("seconds").magnitude),
            comment=None,
            **opts,
        )

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> HoldAndCollect:
        return HoldAndCollect(UR.Quantity(cast(int, sc.args[0]), "seconds"), **sc.opts)  # type: ignore


@dataclass
class Hold(ProtoCommand):
    """A protocol hold (for a time) command."""

    time: pint.Quantity[int] | None
    increment: pint.Quantity[int] = UR.Quantity(0, "seconds")
    incrementcycle: int = 1
    incrementstep: int = 1
    _names: ClassVar[Sequence[str]] = ("HOLD",)

    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
        opts = {}
        if self.increment != Hold.increment:
            opts["increment"] = self.increment.to("seconds").magnitude
        if self.incrementcycle != 1:
            opts["incrementcycle"] = self.incrementcycle
        if self.incrementstep != 1:
            opts["incrementstep"] = self.incrementstep
        return SCPICommand(
            "HOLD",
            int(self.time.to("seconds").magnitude) if self.time is not None else "",
            comment=None,
            **opts,
        )

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> Hold:
        return Hold(UR.Quantity(cast(int, sc.args[0]), "seconds"), **sc.opts)  # type: ignore


class XMLable(ABC):
    @abstractmethod
    def to_xml(self, **kwargs: Any) -> ET.Element:
        ...


G = TypeVar("G")


class _NumOrRefIndexer(Generic[G]):
    _list: list[G]

    def __init__(self, val_list: list[G]):
        self._list = val_list

    @overload
    def _translate_key(self, key: int | str) -> int:
        ...

    @overload
    def _translate_key(self, key: slice) -> slice:
        ...

    def _translate_key(self, key: int | str | slice) -> int | slice:
        if isinstance(key, int):
            return key - 1
        elif isinstance(key, str):
            raise NotImplemented
        elif isinstance(key, slice):
            return slice(key.start - 1, key.stop, key.step)

    def __getitem__(self, key: int | str | slice) -> G | list[G]:
        return self._list[self._translate_key(key)]

    @overload
    def __setitem__(self, key: int | str, val: G) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, val: Sequence[G]) -> None:
        ...

    def __setitem__(self, key, val):
        self._list.__setitem__(self._translate_key(key), val)

    def __delitem__(self, key: int | str | slice) -> None:
        del self._list[self._translate_key(key)]

    def __call__(self, key: int | str | slice) -> G | list[G]:
        return self[key]

    def append(self, val: G) -> None:
        return self._list.append(val)

    def __iadd__(self, val: Iterable[G]) -> None:
        self._list += val


class CustomStep(ProtoCommand, XMLable):
    """A protocol step composed of SCPI/protocol commands."""

    _body: list[ProtoCommand]
    _identifier: int | str | None = None
    repeat: int = 1
    _names: ClassVar[Sequence[str]] = ["STEP"]

    def __init__(
        self,
        body: Sequence[ProtoCommand],
        identifier: int | str | None = None,
        repeat: int = 1,
    ):
        self._body = list(body)
        self._identifier = identifier
        self.repeat = repeat

    def info_str(self, index: None | int = None, repeats: int = 1) -> str:
        if index is not None:
            s = f"{index}. "
        else:
            s = f"- "
        s += f"Step{' '+str(self._identifier) if self._identifier is not None else ''} of commands:\n"
        s += "\n".join(
            f"  {i+1}. " + c.to_scpicommand().to_string()
            for i, c in enumerate(self._body)
        )
        return s

    def duration_at_cycle(self, cycle: int) -> pint.Quantity[int]:  # cycle from 1
        return UR.Quantity(0, "second")

    def temperatures_at_cycle(self, cycle: int) -> pint.Quantity[np.ndarray]:
        return UR.Quantity(np.array(6 * [math.nan]), "degC")

    def total_duration(self, repeat: int = 1) -> pint.Quantity[int]:
        return 0 * UR.seconds

    @property
    def body(self) -> list[ProtoCommand]:
        return self._body

    @body.setter
    def body(self, v: Sequence[ProtoCommand]) -> None:
        self._body = list(v)

    @property
    def identifier(self) -> int | str | None:
        return self._identifier

    @identifier.setter
    def identifier(self, v: int | str | None) -> None:
        self._identifier = v

    @property
    def collects(self) -> bool:
        return False

    def to_xml(self, **kwargs: Any) -> ET.Element:
        assert not kwargs
        e = ET.Element("TCStep")
        ET.SubElement(e, "CollectionFlag").text = str(
            int(self.collects)
        )  # FIXME: approx
        for t in range(0, 6):  # FIXME
            ET.SubElement(e, "Temperature").text = "30.0"
        ET.SubElement(e, "HoldTime").text = "1"
        # FIXME: does not contain cycle starts, because AB format can't handle
        ET.SubElement(e, "ExtTemperature").text = str(0)
        ET.SubElement(e, "ExtHoldTime").text = str(0)
        # FIXME: RampRate, RampRateUnit
        ET.SubElement(e, "RampRate").text = "1.6"
        ET.SubElement(e, "RampRateUnit").text = "DEGREES_PER_SECOND"
        return e

    def to_scpicommand(self, *, stepindex: int = -1, **kwargs: Any) -> SCPICommand:
        opts = {}
        args: list[int | str | None | Sequence[SCPICommand]] = []
        if self.repeat != 1:
            opts["repeat"] = self.repeat
        if self._identifier:
            args.append(self.identifier)
        else:
            args.append(stepindex)

        args.append([com.to_scpicommand(**kwargs) for com in self.body])

        return SCPICommand("STEP", *args, **opts)  # type: ignore

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> CustomStep:
        # Try a normal step:
        try:
            return Step.from_scpicommand(sc)
        except ValueError:
            return cls(
                [
                    cast(ProtoCommand, x.specialize())
                    for x in cast(Sequence[SCPICommand], sc.args[1])
                ],
                identifier=cast(Union[int, str], sc.args[0]),
                **sc.opts,  # type: ignore
            )


def _filterlist(y: Iterable[FilterSet | str]) -> Sequence[FilterSet]:
    return [FilterSet.fromstring(x) for x in y]


@attr.define
class Step(CustomStep, XMLable):
    """
    A normal protocol step, of a hold and possible collection.

    Parameters
    ----------

    time: int
        The step time setting, in seconds.
    temperature: float | Sequence[float]
        The temperature hold setting, either as a float (all zones the same) or a sequence
        (of correct length) of floats setting the temperature for each zone.
    collect
        Collect fluorescence data?  If None (default), collect only if the Step has an explicit
        filters setting.
    temp_increment: float
        Amount to increment all zone temperatures per cycle on and after :any:`temp_incrementcycle`.
    temp_incrementcycle: int (default 2)
        First cycle to start the increment changes. Note that the default in QSLib is 2, not 1 (as in AB's software),
        so that leaving this alone makes sense (the first cycle will be at :any:`temperature`, the next at
        :code:`temperature + temp_incrementcycle`.
    time_increment : float
    time_incrementcycle : int
        The same settings for time per cycle.
    filters : Sequence[FilterSet | str] (default empty)
        A list of filter pairs to collect, either using :any:`FilterSet` or a string like "x1-m4".  If collect is
        True and this is empty, then the filters will be set by the Protocol.

    Notes
    -----

    This currently does not support step-level repeats, which do exist on the machine.
    """

    time: pint.Quantity[int] = attr.field(
        converter=_wrap_seconds, on_setattr=attr.setters.convert
    )
    temperature: pint.Quantity[Any] = attr.field(
        converter=_wrapunitmaybelist_degC, on_setattr=attr.setters.convert
    )
    collect: bool | None = None
    temp_increment: pint.Quantity[float] = attr.field(
        default=UR.Quantity(0.0, UR.delta_degC),
        converter=_wrap_delta_degC,
        on_setattr=attr.setters.convert,
    )
    temp_incrementcycle: int = 2
    temp_incrementpoint: int = 2
    time_increment: pint.Quantity[int] = attr.field(
        default=UR.Quantity(0, UR.second),
        converter=_wrap_seconds,
        on_setattr=attr.setters.convert,
    )
    time_incrementcycle: int = 2
    time_incrementpoint: int = 2
    filters: Sequence[FilterSet] = attr.field(
        default=tuple(),
        converter=_filterlist,
        on_setattr=attr.setters.convert,
    )
    pcr: bool = False
    quant: bool = True
    tiff: bool = False
    repeat: int = 1
    _default_filters: Sequence[FilterSet] = attr.field(default=tuple())
    _classname: ClassVar[str] = "Step"

    def __eq__(self, other: object) -> bool:  # FIXME: add other stuff
        if not isinstance(other, Step):
            return False
        if self.__class__ != other.__class__:
            return False
        if np.all(self.temperature_list != other.temperature_list):
            return False
        # FIXME
        if (self.filters) and (other.filters) and self._filtersets != other._filtersets:
            return False
        if self.time != other.time:
            return False
        if self.collects != other.collects:
            return False
        if self.temp_increment != other.temp_increment:
            return False
        if self.temp_incrementcycle != other.temp_incrementcycle:
            return False
        if self.time_increment != other.time_increment:
            return False
        if self.time_incrementcycle != other.time_incrementcycle:
            return False
        if self.pcr != other.pcr:
            return False
        if self.quant != other.quant:
            return False
        if self.tiff != other.tiff:
            return False
        return True

    @property
    def collects(self):
        if self.collect is None:
            return len(self.filters) > 0
        return self.collect

    @property
    def _filtersets(self):
        return [FilterSet.fromstring(x) for x in self.filters]

    def info_str(self, index: None | int = None, repeats: int = 1) -> str:
        "String describing the step."

        temperatures_cycle1 = self.temperatures_at_cycle(1)
        temperatures_cycle1 = convert_quantity_ndarray_to_scalar_if_all_equal(
            temperatures_cycle1
        )
        tempstr = "{:.2f~}".format(temperatures_cycle1)  # type: ignore
        if (repeats > 1) and (self.temp_increment != 0.0):
            temperatures = self.temperatures_at_cycle(repeats)
            temperatures = convert_quantity_ndarray_to_scalar_if_all_equal(temperatures)
            t = "{:.2f~}".format(temperatures)  # type: ignore
            tempstr += f" to {t}"

        elems = [f"{tempstr} for {self.time}/cycle"]
        if self.temp_increment != 0.0:
            elems.append(f"{self.temp_increment}°C/cycle")
            if self.temp_incrementcycle > 1:
                elems[-1] += f" from cycle {self.temp_incrementcycle}"
        if self.time_increment != 0.0:
            elems.append(f"{_durformat(self.time_increment)}/cycle")
            if self.time_incrementcycle != 2:
                elems[
                    -1
                ] += f" from cycle {self.time_incrementcycle}"  # Fixme: ADD OTHER STUFF
        # if self.ramp_rate != 1.6:
        #    elems.append(f"{self.ramp_rate} °C/s ramp")
        s = f"{index}. " + ", ".join(elems)

        if self.collects:
            s += " (collects "
            if self.filters:
                s += ", ".join(FilterSet.fromstring(f).lowerform for f in self.filters)
            else:
                s += "default"
            if self.pcr:
                s += ", pcr on"
            if not self.quant:
                s += ", quant off"
            if self.tiff:
                s += ", keeps images"
            s += ")"

        return s

    def total_duration(self, repeats: int = 1) -> pint.Quantity:
        return sum(
            (self.duration_at_cycle(c) for c in range(1, repeats + 1)), 0 * UR.seconds
        )

    def duration_at_cycle(
        self, cycle: int
    ) -> pint.Quantity:  # cycle from 1 # FIXME: add point
        "Duration of the step (excluding ramp) at `cycle` (from 1)"
        inccycles = max(0, cycle + 1 - self.time_incrementcycle)
        return self.time + inccycles * self.time_increment
        # FIXME: is this right?

    def temperatures_at_cycle(
        self, cycle: int
    ) -> pint.Quantity[np.ndarray]:  # FIXME: add point
        "Temperatures of the step at `cycle` (from 1)"
        inccycles = max(0, cycle + 1 - self.temp_incrementcycle)
        return self.temperature_list + inccycles * self.temp_increment

    @property
    def identifier(self) -> None:
        return None

    @identifier.setter
    def identifier(self, v: Any) -> None:
        raise ValueError

    @property
    def temperature_list(self) -> pint.Quantity[np.ndarray]:
        mag = self.temperature.to("degC").magnitude  # FIXME
        if isinstance(mag, np.ndarray):
            return UR.Quantity(mag, "degC")
        else:
            return UR.Quantity(NZONES * [mag], "degC")

    @property
    def body(self) -> list[ProtoCommand]:
        if self.collects:
            return [
                Ramp(
                    self.temperature_list,
                    self.temp_increment,
                    self.temp_incrementcycle,
                    self.temp_incrementpoint,
                ),
                HACFILT(self.filters),
                HoldAndCollect(
                    self.time,
                    self.time_increment,
                    self.time_incrementcycle,
                    self.time_incrementpoint,
                    self.tiff,
                    self.quant,
                    self.pcr,
                ),
            ]
        else:
            return [
                Ramp(
                    self.temperature_list,
                    self.temp_increment,
                    self.temp_incrementcycle,
                    self.temp_incrementpoint,
                ),
                Hold(
                    self.time,
                    self.time_increment,
                    self.time_incrementcycle,
                    self.time_incrementpoint,
                ),
            ]

    @body.setter
    def body(self, v: Any) -> None:
        raise ValueError

    @classmethod
    def from_xml(
        cls, e: ET.Element, *, etc: int = 1, ehtc: int = 1, he: bool = False
    ) -> Step:
        collect = bool(int(e.findtext("CollectionFlag") or 0))
        ts: pint.Quantity[np.ndarray] = UR.Quantity(
            [float(x.text or math.nan) for x in e.findall("Temperature")], "degC"
        )
        ht: pint.Quantity[int] = int(e.findtext("HoldTime") or 0) * UR("seconds")
        et: pint.Quantity[float] = float(e.findtext("ExtTemperature") or 0.0) * UR(
            "delta_degC"
        )
        eht: pint.Quantity[int] = int(e.findtext("ExtHoldTime") or 0) * UR("seconds")
        if not he:
            et = 0 * UR("seconds")
            eht = 0 * UR("seconds")
        return Step(ht, ts, collect, et, etc, 1, eht, ehtc, 1, [], True)

    def to_xml(self, **kwargs: Any) -> ET.Element:
        assert not kwargs
        e = ET.Element("TCStep")
        ET.SubElement(e, "CollectionFlag").text = str(
            int(self.collects)
        )  # FIXME: approx
        for t in self.temperature_list.to("°C").magnitude:
            ET.SubElement(e, "Temperature").text = str(t)
        ET.SubElement(e, "HoldTime").text = str(int(self.time.to("seconds").magnitude))
        # FIXME: does not contain cycle starts, because AB format can't handle
        ET.SubElement(e, "ExtTemperature").text = str(
            self.temp_increment.to("delta_degC").magnitude
        )
        ET.SubElement(e, "ExtHoldTime").text = str(
            int(self.time_increment.to("seconds").magnitude)
        )
        # FIXME: RampRate, RampRateUnit
        ET.SubElement(e, "RampRate").text = "1.6"
        ET.SubElement(e, "RampRateUnit").text = "DEGREES_PER_SECOND"
        return e

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand) -> Step:
        coms = [x.specialize() for x in cast(Sequence[SCPICommand], sc.args[1])]

        # FIXME: When willing to require Python 3.10, this can be cleaned up with a match statement.

        h: Hold | HoldAndCollect

        com_classes = [x.__class__ for x in coms]
        if com_classes == [Ramp, HACFILT, HoldAndCollect]:
            r = cast(Ramp, coms[0])
            hcf = cast(HACFILT, coms[1])
            h = cast(HoldAndCollect, coms[2])
            c = cls(
                h.time,
                r.temperature,
                time_incrementcycle=1,
                temp_incrementcycle=1,
            )
            c.collect = True
            if hcf._default_filters:
                c.filters = []
                c.collect = True
                c._default_filters = hcf._default_filters
            else:
                c.filters = hcf.filters
            c.time_increment = h.increment
            c.temp_increment = r.increment
            c.time_incrementcycle = h.incrementcycle
            c.temp_incrementcycle = r.incrementcycle
        elif com_classes == [Ramp, Hold]:
            r = cast(Ramp, coms[0])
            h = cast(Hold, coms[1])
            if h.time is None:
                raise ValueError
            c = cls(
                h.time,
                r.temperature,
                time_incrementcycle=1,
                temp_incrementcycle=1,
            )
            c.collect = False
            c.time_increment = h.increment
            c.temp_increment = r.increment
            c.time_incrementcycle = h.incrementcycle
            c.temp_incrementcycle = r.incrementcycle
        else:
            raise ValueError
        return c

    @classmethod
    def fromdict(cls, d: dict[str, Any]) -> "Step":
        return cls(**d)


def convert_quantity_ndarray_to_scalar_if_all_equal(
    quants: pint.Quantity,
) -> pint.Quantity:
    """
    If `quants` is a `Quantity[ndarray]`, but all floats in the ndarray are exactly equal,
    then return a `Quantity[unit]`, where `unit` is the unit of `quants` (e.g., degC).

    :param quants:
        Quantity[ndarray]
    :return:
        `quants` unchanged if the array has more than one value,
        otherwise a Quantity with the single shared value
    """
    if len(set(quants.m)) == 1:
        temperatures_cycle1_float = quants.m[0]
        quants = pint.Quantity(temperatures_cycle1_float, quants.u)
    return quants


def _bsl(x: Iterable[CustomStep] | CustomStep) -> Sequence[CustomStep]:
    return list(x) if isinstance(x, Iterable) else [x]


def _maybelist_cs(x: CustomStep | Sequence[CustomStep]) -> Sequence[CustomStep]:
    return [x] if not isinstance(x, Sequence) else list(x)


@attr.define
class Stage(XMLable, ProtoCommand):
    """A Stage in a protocol, composed of :class:`Step` s with a possible repeat."""

    steps: Sequence[CustomStep] = attr.field(converter=_maybelist_cs)
    repeat: int = 1
    index: int | None = None
    label: str | None = None
    _default_filters: Sequence[FilterSet] = attr.field(default=tuple())
    _classname: ClassVar[str] = "Stage"
    _names: ClassVar[Sequence[str]] = ("STAGe",)

    @property
    def step(self) -> _NumOrRefIndexer[CustomStep]:
        return _NumOrRefIndexer(list(self.steps))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stage):
            return False
        if self.__class__ != other.__class__:
            return False
        if (
            (self.index is not None)
            and (other.index is not None)
            and self.index != other.index
        ):
            return False
        if (
            (self.label is not None)
            and (other.label is not None)
            and self.label != other.label
        ):
            return False
        return self.steps == other.steps

    @classmethod
    def stepped_ramp(
        cls: Type[Stage],
        from_temperature: float | str | pint.Quantity[float] | Sequence[float],
        to_temperature: float | str | pint.Quantity[float] | Sequence[float],
        total_time: int | str | pint.Quantity[int],
        *,
        n_steps: int | None = None,
        temperature_step: float | str | pint.Quantity[float] | None = None,
        collect: bool | None = None,
        filters: Sequence[str | FilterSet] = tuple(),
    ) -> Stage:
        """Hold at a series of temperatures, from one to another.

        Parameters
        ----------
        from_temperature
            Initial temperature/s (inclusive).
        to_temperature
            Final temperature/s (inclusive).
        total_time
            Total time for the stage
        n_steps
            Number of steps.  If None, uses 1.0 Δ°C steps, or, if
            doing a multi-temperature change, uses maximum step
            of 1.0 Δ°C.
        temperature_step
            Step temperature change (optional).  Must be None,
            or correctly match calculation, if n_steps is not
            None.  If both this and n_steps are None, default
            is 1.0 Δ°C steps.  If temperature step does not
            exactly fit range, it will be adjusted, with a warning
            if the change is more than 5%.  Sign is ignored.  If
            doing a multi-temperature change, then this is the
            maximum temperature step.
        collect
            Collect data?  If None, collects data if filters is set explicitly.
        filters
            Filters to collect.

        Returns
        -------
        Stage
            The resulting stage.
        """

        from_temperature = _wrapunitmaybelist_degC(from_temperature)
        to_temperature = _wrapunitmaybelist_degC(to_temperature)

        delta = to_temperature - from_temperature

        multistep = False

        if hasattr(delta, "shape"):
            if (delta != delta[0]).any():
                multistep = True
                max_delta = delta.max()  # type: pint.Quantity[float]
            else:
                max_delta = delta[0]
        else:
            max_delta = delta

        total_time = _wrap_seconds(total_time).to("seconds")

        if n_steps is None:
            if temperature_step is None:
                temperature_step = UR("1.0 delta_degC")
            else:
                temperature_step = abs(_wrap_delta_degC(temperature_step))

            n_steps = abs(round((max_delta / temperature_step).to("").magnitude)) + 1

            real_max_temperature_step = abs(max_delta / (n_steps - 1))

            change = (
                ((real_max_temperature_step - temperature_step) / temperature_step)
                .to("")
                .magnitude
            )

            if abs(change) > 0.05:
                warnings.warn(
                    f"Desired temperature step {temperature_step} differs by {100*change}% from actual {real_max_temperature_step}."
                )

        elif temperature_step is not None:
            temperature_step = abs(_wrap_delta_degC(temperature_step))
            if (
                abs(round((max_delta / temperature_step).to("").magnitude)) + 1
                != n_steps
            ):
                raise ValueError(
                    "Both n_steps and temperature_step set, and calculated steps don't match set steps."
                )

        temp_increment = ((to_temperature - from_temperature) / (n_steps - 1)).round(4)

        step_time = (total_time / n_steps).round()

        if not multistep:
            return cls(
                [
                    Step(
                        step_time,
                        from_temperature,
                        collect=collect,
                        temp_increment=temp_increment,
                        filters=filters,
                    )
                ],
                repeat=n_steps,
            )

        # MULTISTEP!

        return cls(
            [
                Step(
                    step_time,
                    from_temperature + step_i * temp_increment,
                    collect=collect,
                    filters=filters,
                )
                for step_i in range(0, n_steps)
            ]
        )

    @classmethod
    def hold_at(
        cls: Type[Stage],
        temperature: float | str | Sequence[float],
        total_time: int | str | pint.Quantity[int],
        step_time: int | str | pint.Quantity[int] | None = None,
        collect: bool | None = None,
        filters: Sequence[str | FilterSet] = tuple(),
    ) -> Stage:
        """Hold at a temperature for a set amount of time, with steps of a configurable fixed time.


        Parameters
        ----------
        temperatures
            The temperature or temperatures to hold.  If not strings or quantities, the value/values are
            interpreted as °C.
        total_time
            Desired total time for the stage.  If this is not a multiple of `step_time`, it may not be the actual
            total time for the stage.  The function will emit a warning if the difference is more than 10%.  If
            an integer, value is interpreted as seconds.
        step_time
            If None (default), the stage will have one step.  Otherwise, it will have steps of this time.  If
            an integer, value is interpreted as seconds.
        collect
            Whether or not each step should collect fluorescence data.  If None (default), collects data
            if filters is explicitly set.
        filters
            A list of filters to collect.  If empty, and collect is True, then each step will collect the
            default filters for the :class:`Protocol`.

        Returns
        -------
        Stage
            The resulting Stage

        Raises
        ------
        ValueError
            If step time is larger than total time.
        """
        if step_time is None:
            step_time = total_time

        step_time = _wrap_seconds(step_time)
        total_time = _wrap_seconds(total_time)

        if step_time > total_time:
            raise ValueError(
                f"Step time {step_time} > total time {total_time}.  Did you mix up the parameter order?"
            )

        repeat = round((total_time / step_time).to("").magnitude)

        real_total_time = repeat * step_time

        chg = ((real_total_time - total_time) / total_time).to("").magnitude

        if abs(chg) > 0.1:
            warnings.warn(
                f"Stage will have total time {real_total_time}, with {repeat} steps, {100*chg} different from desired time {total_time}."
            )

        return cls(
            [
                Step(
                    step_time,
                    _wrapunitmaybelist_degC(temperature),
                    collect=collect,
                    filters=filters,
                )
            ],
            repeat=repeat,
        )

    hold_for = hold_at

    def __repr__(self) -> str:
        s = f"Stage(steps="
        if len(self.steps) == 0:
            s += "[]"
        elif len(self.steps) == 1:
            s += repr(self.steps[0])
        else:
            s += "[" + ", ".join(repr(f) for f in self.steps) + "]"
        if self.repeat != 1:
            s += f", repeat={self.repeat}"
        if self.index:
            s += f", index={self.index}"
        if self.label:
            s += f", label={self.label}"
        s += ")"
        return s

    def dataframe(
        self, start_time: float = 0, previous_temperatures: list[float] | None = None
    ) -> pd.DataFrame:
        """
        Create a dataframe of the steps in this stage.

        Parameters
        ----------

        start_time
            The initial start time, in seconds, of the stage (before
            the ramp to the first step).  Default is 0.

        previous_temperatures
            A list of temperatures at the end of the previous stage, to allow
            calculation of ramp time.  If `None`, the ramp is assumed to take
            no time.
        """

        durations = np.array(
            [
                step.duration_at_cycle(i).to("seconds").magnitude
                for i in range(1, self.repeat + 1)
                for step in self.steps
            ]
        )
        temperatures = np.array(
            [
                step.temperatures_at_cycle(i).to("°C").magnitude
                for i in range(1, self.repeat + 1)
                for step in self.steps
            ]
        )
        ramp_rates = [1.6 for _ in range(1, self.repeat + 1) for step in self.steps]
        # np.array(
        #    [step.ramp_rate for _ in range(1, self.repeat + 1) for step in self.body]
        # )
        collect_data = np.array(
            [step.collects for _ in range(1, self.repeat + 1) for step in self.steps]
        )

        # FIXME: is this how ramp rates actually work?

        ramp_durations = np.zeros(len(durations))
        if previous_temperatures is not None:
            ramp_durations[0] = (
                np.max(np.abs(temperatures[0] - previous_temperatures)) / ramp_rates[0]
            )
            ramp_durations[1:] = (
                np.max(np.abs(temperatures[1:] - temperatures[:-1]), axis=1)
                / ramp_rates[1:]
            )

        tot_durations = durations + ramp_durations

        start_times = start_time + np.zeros(len(durations))
        start_times[0] = start_time + ramp_durations[0]
        start_times[1:] = (
            start_time + np.cumsum(tot_durations[:-1]) + ramp_durations[1:]
        )

        end_times = start_time + np.cumsum(tot_durations)

        data = pd.DataFrame(
            {
                "start_time": start_times,
                "end_time": end_times,
                "collect_data": collect_data,
            }
        )

        data["temperature_avg"] = np.average(temperatures, axis=1)

        for i in range(
            0, temperatures.shape[1]
        ):  # pylint: disable=unsubscriptable-object
            data["temperature_{}".format(i + 1)] = temperatures[:, i]

        data["cycle"] = [
            c for c in range(1, self.repeat + 1) for s in range(1, len(self.steps) + 1)
        ]
        data["step"] = [
            s for c in range(1, self.repeat + 1) for s in range(1, len(self.steps) + 1)
        ]

        data.set_index(["cycle", "step"], inplace=True)

        return data

    def to_scpicommand(
        self, stageindex: int | str | None = None, **kwargs: Any
    ) -> SCPICommand:
        opts = {}
        args: list[int | str | list[SCPICommand]] = []
        if self.repeat != 1:
            opts["repeat"] = self.repeat
        if self.index:
            index_to_use: int | str = self.index
        elif stageindex:
            index_to_use = stageindex
        else:
            raise ValueError("No index.")
        args.append(index_to_use)
        args.append(self.label or f"STAGE_{index_to_use}")
        args.append(
            [
                step.to_scpicommand(stepindex=i + 1, **kwargs)
                for i, step in enumerate(self.steps)
            ]
        )

        return SCPICommand("STAGe", *args, comment=None, **opts)

    @classmethod
    def from_scpicommand(cls, sc: SCPICommand, **kwargs: Any) -> Stage:
        c = cls(
            [
                cast(CustomStep, x.specialize(**kwargs))
                for x in cast(Sequence[SCPICommand], sc.args[2])
            ],
            index=cast(int, sc.args[0]),
            label=cast(Optional[str], sc.args[1]),
            **sc.opts,  # type: ignore
        )

        dfilt: list[FilterSet] = []

        for s in c.steps:
            if isinstance(s, Step) and s._default_filters:
                ndf = s._default_filters
                if len(dfilt) == 0:
                    dfilt = list(ndf)
                if dfilt != ndf:
                    raise ValueError("Inconsistent default filters")
        c._default_filters = dfilt

        return c

    @classmethod
    def from_xml(cls, e: ET.Element) -> Stage:
        rep = int(e.findtext("NumOfRepetitions") or 1)
        startcycle = int(cast(str, e.findtext("StartingCycle")))
        ade = e.findtext("AutoDeltaEnabled") == "true"
        steps: list[CustomStep] = [
            Step.from_xml(x, etc=startcycle, ehtc=startcycle, he=ade)
            for x in e.findall("TCStep")
        ]
        return cls(steps, rep)

    def to_xml(self, **kwargs: Any) -> ET.Element:
        assert not kwargs
        e = ET.Element("TCStage")
        ET.SubElement(e, "StageFlag").text = "CYCLING"
        ET.SubElement(e, "NumOfRepetitions").text = str(int(self.repeat))
        scycle: set[int] = set()
        for s in self.steps:
            if isinstance(s, Step):
                scycle.add(s.temp_incrementcycle)
                scycle.add(s.time_incrementcycle)
            if isinstance(s, XMLable):
                e.append(s.to_xml())
        if len(scycle) > 1:
            log.warn("Approx")
        if scycle:
            ET.SubElement(e, "StartingCycle").text = str(next(iter(scycle)))
        ET.SubElement(e, "AutoDeltaEnabled").text = "true"

        return e

    def info_str(self, index: int | None = None) -> str:
        if self.repeat > 1:
            adds = "s"
        else:
            adds = ""
        stagestr = f"{index}. Stage with {self.repeat} cycle{adds}"
        stepstrs = [
            textwrap.indent(f"{step.info_str(i+1, self.repeat)}", "    ")
            for i, step in enumerate(self.steps)
        ]
        try:
            tot_dur = sum(
                (x.total_duration(self.repeat) for x in self.steps),
                UR.Quantity(0, UR.seconds),
            )
            stagestr += f" (total duration {_durformat(tot_dur)})"
        except KeyError:
            pass
        if len(stepstrs) > 1:
            stagestr += " of:\n" + "\n".join(stepstrs)
        else:
            stagestr += " of " + " ".join(stepstrs[0].split()[1:])
        return stagestr

    def __str__(self) -> str:
        return self.info_str(None)

    @classmethod
    def fromdict(cls, d: dict[str, Any]) -> "Stage":
        s = [cast(CustomStep, Step(**x)) for x in d["_steps"]]
        del d["_steps"]
        return cls(s, **d)


def _oxfordlist(iterable: Iterable[str]) -> str:
    x = iter(iterable)
    try:
        s = next(x)  # we know the first will be there
    except StopIteration:
        return ""
    try:
        maybeult = next(x)
    except StopIteration:
        return s
    try:
        nextult = next(x)
        s = s + ", " + maybeult
        maybeult = nextult
    except StopIteration:
        return s + " and " + maybeult
    while True:
        try:
            nextult = next(x)
            s = s + ", " + maybeult
            maybeult = nextult
        except StopIteration:
            return s + ", and " + maybeult


ALLTEMPS = ["temperature_{}".format(i) for i in range(1, 7)]


@attr.define
class Protocol(ProtoCommand):
    """A run protocol for the QuantStudio.  Protocols encapsulate the temperature and camera
    controls for an entire run.  They are composed of :class:`Stage`s, which may repeat for a
    number of cycles, and the stages are in turn composed of Steps, which may be created for
    usual cases with :class:`Step`, or from SCPI commands.  Steps may repeat
    their contents as well, but this is not yet implemeted.

    Parameters
    ----------
    stages: Iterable[Stage]
        The stages of the protocol, likely :class:`Stage`.
    stage: _NumOrRefIndexer[Stage]
        A more convenient way of accessing the stages of the protocol, with
        numbering that matches the machine.
    name: str | None
        A protocol name. If not set, a timestamp will be used, unlike AB's uuid.
    volume: float
        The sample volume, in µL.
    runmode: str | None
        The run mode.
    covertemperature: float (default 105.0)
        The cover temperature
    filters: Sequence[str]
        A list of default filters that can be used by any collection commands
        that don't specify their own.
    prerun: Sequence[SCPICommand]
        Sets PRERUN.  *DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING*.
    postrun: Sequence[SCPICommand]
        Sets POSTRUN. *DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING*.
    """

    stages: list[Stage] = attr.field(factory=list)
    name: str = attr.field(factory=lambda: "Prot_" + _nowuuid())
    volume: float = 50.0
    runmode: str = "standard"  # standard or fast... need to deal with this
    filters: Sequence[FilterSet] = attr.field(
        factory=lambda: [],
        converter=_filterlist,
        on_setattr=attr.setters.convert,
    )
    covertemperature: float = 105.0
    prerun: Sequence[SCPICommandLike] = attr.field(factory=list)
    postrun: Sequence[SCPICommandLike] = attr.field(factory=list)
    _classname: str = "Protocol"
    _names: ClassVar[Sequence[str]] = ("PROTocol", "PROT")

    @property
    def stage(self) -> _NumOrRefIndexer[Stage]:
        """
        A more convenient view of :any:`Protocol.stages`.  This allows one-indexed access,
        such that `protocol.stage[5] == protocol.stages[6]` is stage 5 of the protocol, in
        the interpretation of tha machine.  Indexing can use slices, and is *inclusive*, so
        `protocol.stage[5:6]` returns stages 5 and 6.  Getting, setting, and appending
        stages are all supported through this interface.
        """
        return _NumOrRefIndexer(self.stages)

    def to_scpicommand(self, **kwargs: Any) -> SCPICommand:
        assert not kwargs
        args: list[int | str | Sequence[SCPICommand]] = []
        opts: dict[str, int | float | str] = {}
        if self.volume is not None:
            opts["volume"] = self.volume
        if self.runmode is not None:
            opts["runmode"] = self.runmode
        args.append(self.name)
        stages: list[SCPICommand] = []
        if self.prerun:
            stages.append(
                SCPICommand("PRERun", [s.to_scpicommand() for s in self.prerun])
            )

        stages += [
            stage.to_scpicommand(
                filters=self.filters, stageindex=i + 1, default_filters=self.filters
            )
            for i, stage in enumerate(self.stages)
        ]

        if self.postrun:
            stages.append(
                SCPICommand("POSTRun", [s.to_scpicommand() for s in self.postrun])
            )

        args.append(stages)

        return SCPICommand("PROTocol", *args, comment=None, **opts)

    @classmethod
    def from_scpicommand(cls: Type[Protocol], sc: SCPICommand) -> Protocol:
        stage_commands = [x for x in cast(Sequence[SCPICommand], sc.args[1])]

        if stage_commands[0].command.upper() == "PRERUN":
            prerun = cast(Sequence[SCPICommand], stage_commands[0].args[0])
            assert isinstance(prerun, Sequence)
            del stage_commands[0]
        else:
            prerun = []
        if stage_commands[-1].command.upper() == "POSTRUN":
            postrun = cast(Sequence[SCPICommand], stage_commands[-1].args[0])
            assert isinstance(postrun, Sequence)
            del stage_commands[-1]
        else:
            postrun = []

        stages = [cast(Stage, x.specialize()) for x in stage_commands]
        for s in stages:
            if not isinstance(s, Stage):
                raise ValueError

        name = sc.args[0]
        if not isinstance(name, str):
            raise ValueError

        opts = sc.opts

        c = cls(stages, name, prerun=prerun, postrun=postrun, **opts)  # type: ignore

        dfilt: list[FilterSet] = []

        for s in c.stages:
            if hasattr(s, "_default_filters") and s._default_filters:
                ndf = s._default_filters
                if len(dfilt) == 0:
                    dfilt = list(ndf)
                if dfilt != ndf:
                    raise ValueError("Inconsistent default filters")
        c.filters = dfilt

        return c

    @property
    def dataframe(self) -> pd.DataFrame:
        "A DataFrame of the temperature protocol."
        dataframes: list[pd.DataFrame] = []

        stage_start_time = 0.0
        stagenum = 1
        previous_temperatures = [25.0] * 6

        for stage in self.stages:
            dataframe = stage.dataframe(stage_start_time, list(previous_temperatures))
            stagenum += 1
            previous_temperatures = dataframe.iloc[-1].loc[ALLTEMPS].to_numpy()
            stage_start_time = cast(float, dataframe["end_time"].iloc[-1])
            dataframes.append(dataframe)

        return pd.concat(
            dataframes,
            keys=range(1, len(dataframes) + 1),
            names=["stage", "step", "cycle"],
        )

    @property
    def all_points(self) -> pd.DataFrame:
        d = self.dataframe
        dd = (
            d.loc[:, "temperature_avg":]  # type: ignore
            .iloc[np.repeat(np.arange(0, len(d)), 2)]
            .reset_index()
        )
        times = np.empty(len(dd))
        times[::2] = np.array(d.loc[:, "start_time"])
        times[1::2] = np.array(d.loc[:, "end_time"])
        dd.insert(0, "time", times)
        return dd

    @property
    def all_times(self) -> np.ndarray:
        "An array of all start and end times of each step, interleaved."
        d = self.dataframe
        alltimes = np.empty(len(d) * 2)
        alltimes[:-1:2] = d["start_time"]
        alltimes[1::2] = d["end_time"]
        return alltimes

    @property
    def all_filters(self) -> Collection[FilterSet]:
        "A list of all filters used at some point in the protocol."
        filters = {FilterSet.fromstring(f) for f in self.filters}

        for stage in self.stages:
            for step in stage.steps:
                if isinstance(step, Step) and step.collect:
                    filters |= {FilterSet.fromstring(f) for f in step.filters}

        return filters

    def copy(self) -> Protocol:
        """Returns a new copy (recursively) of the protocol."""
        return deepcopy(self)

    @property
    def all_temperatures(self) -> np.ndarray:
        "An array of temperature settings at `all_times`."
        d = self.dataframe
        return np.repeat(d["temperature_avg":], 2)  # type: ignore

    def plot_protocol(
        self, ax: Optional[plt.Axes] = None
    ) -> Tuple[plt.Axes, Tuple[List[plt.Line2D], List[plt.Line2D]]]:
        "A plot of the temperature and data collection points."

        import matplotlib.pyplot as plt

        #    try:
        #    except ModuleNotFoundError:
        #        raise "

        if ax is None:
            _, ax = plt.subplots()

        d = self.dataframe

        all_points = self.all_points

        # FIXME: 0 to 6 is specific
        p1 = [
            ax.plot(
                all_points["time"] / 3600.0,
                all_points[f"temperature_{i}"],
                label=f"{i}",
            )[0]
            for i in range(1, 7)
        ]

        dc = d.loc[d.collect_data, :]
        p2 = [
            ax.plot(
                cast("pd.Series['float']", dc["end_time"]) / 3600.0,
                dc[f"temperature_{i+1}"],
                ".",
                color=p.get_color(),
            )[0]
            for i, p in enumerate(p1)
        ]

        ax.legend(title="zone")
        ax.set_xlabel("experiment time (hrs) (points collect data)")
        ax.set_ylabel("temperature (°C)")
        ax.set_title("Temperature Protocol")

        return (ax, (p1, p2))

    @classmethod
    def from_xml(cls, e: ET.Element) -> Protocol:
        svol = float(e.findtext("SampleVolume") or 50.0)
        runmode = e.findtext("RunMode") or "standard"
        covertemperature = float(e.findtext("CoverTemperature") or 105.0)
        protoname = e.findtext("ProtocolName")
        if protoname is None:
            raise ValueError
        filter_e = e.find("CollectionProfile")
        filters = []
        if filter_e:
            for x in filter_e.findall("CollectionCondition/FilterSet"):
                filters.append(x.attrib["Excitation"] + "-" + x.attrib["Emission"])
        stages = [Stage.from_xml(x) for x in e.findall("TCStage")]
        return Protocol(stages, protoname, svol, runmode, filters, covertemperature)

    def to_xml(
        self, covertemperature: float = 105.0
    ) -> tuple[ET.ElementTree, ET.ElementTree]:
        te = ET.ElementTree(ET.Element("TCProtocol"))
        tqe = ET.ElementTree(ET.Element("QSTCProtocol"))

        e = te.getroot()
        qe = tqe.getroot()

        _set_or_create(e, "FileVersion").text = "2.0"
        _set_or_create(e, "ProtocolName").text = self.name
        _set_or_create(qe, "QSLibNote").text = (
            "This protocol was"
            " generated by QSLib. It may be only an approximation or"
            " placeholder for the real protocol, contained as"
            " an SCPI command in QSLibProtocolCommand."
        )
        _set_or_create(
            qe, "QSLibProtocolCommand"
        ).text = self.to_scpicommand().to_string()
        _set_or_create(qe, "QSLibProtocol").text = str(attr.asdict(self))
        _set_or_create(qe, "QSLibVerson").text = __version__
        _set_or_create(e, "CoverTemperature").text = str(covertemperature)
        _set_or_create(e, "SampleVolume").text = str(self.volume or 50)
        _set_or_create(e, "RunMode").text = str(self.runmode or "Standard")
        _set_or_create(e, "UserName").text = ""
        _set_or_create(e, "TubeType").text = "0"
        _set_or_create(e, "BlockID").text = "18"  # FIXME
        _set_or_create(e, "Delay").text = "0.0"
        _set_or_create(e, "ExtendedPCRCycles").text = "0"
        _set_or_create(e, "ExtendedHoldTemp").text = "0"
        _set_or_create(e, "ExtendedHoldTime").text = "0"
        if self.filters:
            x = _set_or_create(e, "CollectionProfile", ProfileId="1")
            for f in self.filters:
                if not isinstance(f, FilterSet):
                    f = FilterSet.fromstring(f)
                x.append(f.to_xml())
        for s in self.stages:
            e.append(s.to_xml())
        return te, tqe

    def info(self) -> str:
        """Generate a (markdown) text protocol description."""
        return str(self)

    def __str__(self) -> str:
        begin = f"Run Protocol {self.name}"
        extras = []
        if self.volume:
            extras.append(f"sample volume {self.volume} µL")
        if self.runmode:
            extras.append(f"run mode {self.runmode}")
        if extras:
            begin += " with " + _oxfordlist(extras)
        begin += ":\n"
        if self.filters:
            begin += (
                f"(default filters "
                + _oxfordlist(FilterSet.fromstring(f).lowerform for f in self.filters)
                + ")\n\n"
            )
        else:
            begin += "\n"
        stagestrs = [
            textwrap.indent(stage.info_str(i + 1), "  ")
            for i, stage in enumerate(self.stages)
        ]

        return begin + "\n".join(stagestrs)

    def _repr_markdown_(self) -> str:
        return str(self)

    @classmethod
    def fromdict(cls, d: dict[str, Any]) -> "Protocol":
        s = [Stage(**x) for x in d["stages"]]
        del d["stages"]
        return cls(s, **d)

    def check_compatible(self, new: Protocol, status: RunStatus) -> bool:
        """Checks compatibility for changing a running protocol to a new one.

        Raises ValueError if incompatible, returns True if compatible.

        Parameters
        ----------
        new
            New protocol.
        status
            Current run status.

        Raises
        ------
        ValueError
            Protocols are incompatible.
        """
        # Sample sample volume? (FIXME: We can't change it right now.)
        # assert self.volume == new.volume
        # assert self.name == new.name

        for i, (oldstage, newstage) in enumerate(zip_longest(self.stages, new.stages)):
            if (
                i + 1 < status.stage
            ):  # If the stage has already passed, we must be equal
                if oldstage != newstage:
                    raise ValueError
            elif (
                i + 1 == status.stage
            ):  # Current stage.  Only change is # cycles, >= current
                if newstage.repeat < status.cycle:
                    raise ValueError
                oldstage.repeat = newstage.repeat  # for comparison
                if oldstage != newstage:
                    raise ValueError
            else:
                continue

        return True

    def validate(self, fix: bool = True):
        for i, stage in enumerate(self.stages):
            if (stage.index is not None) and (stage.index != i + 1):
                if fix:
                    log.warn(
                        "Stage %s is at index %d of protocol, but has set index %d. Fixing.",
                        stage,
                        i + 1,
                        stage.index,
                    )
                    stage.index = i + 1
                else:
                    raise ValueError(
                        "Stage %s is at index %d of protocol, but has set index %d."
                        % (stage, i + 1, stage.index)
                    )

            if stage.label is not None:
                m = re.match(r"STAGE_(\d+)", stage.label)
                if m and (int(m[1]) != i + 1):
                    if fix:
                        log.warn(
                            "Stage %s has label %s, which implies index %d, but is at index %d of protocol. Fixing.",
                            stage,
                            stage.label,
                            int(m[1]),
                            i + 1,
                        )
                        stage.label = f"STAGE_{i+1}"
                    else:
                        raise ValueError(
                            "Stage %s has label %s, which implies index %d, but is at index %d of protocol."
                            % (stage, stage.label, int(m[1]), i + 1)
                        )


for c in cast(
    Sequence[Protocol],
    [Ramp, Exposure, HACFILT, HoldAndCollect, Hold, CustomStep, Stage, Protocol],
):
    for n in c._names:
        scpi_commands._scpi_command_classes[n.upper()] = cast(SCPICommandLike, c)



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

"""Experiment class and related.
"""
from __future__ import annotations

import base64
import io
import logging
import os
import re
import tempfile
import xml.etree.ElementTree as ET
import zipfile
from dataclasses import InitVar, dataclass, field
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
from warnings import warn

import numpy as np
import pandas as pd
import toml as toml
from pandas.core.base import DataError

from qslib.plate_setup import PlateSetup
from qslib.scpi_commands import AccessLevel, SCPICommand

from ._analysis_protocol_text import _ANALYSIS_PROTOCOL_TEXT
from ._util import _nowuuid, _pp_seqsliceint, _set_or_create
from .base import RunStatus
from .data import FilterDataReading, FilterSet, df_from_readings
from .machine import Machine
from .processors import NormRaw, Processor
from .protocol import Protocol, Stage, Step
from .rawquant_compat import _fdc_to_rawdata
from .version import __version__

if TYPE_CHECKING:  # pragma: no cover
    import matplotlib.pyplot as plt


TEMPLATE_NAME = "ruo"

_MANIFEST_CONTENTS = f"""Manifest-Version: 1.0
Content-Type: Std
Implementation-Title: QSLib
Implementation-Version: {__version__}
Specification-Vendor: Applied Biosystems
Specification-Title: Experiment Document Specification
Specification-Version: 1.3.2
"""

log = logging.getLogger(__name__)


MachineReference = Union[str, Machine]


@dataclass
class MachineError(Exception):
    """Base class for an error from a machine."""

    host: str = field(init=False)
    port: int | str = field(init=False)
    machine: InitVar[Machine]

    def __post_init__(self, machine: Machine) -> None:
        self.host = machine.host
        self.port = machine.port


@dataclass
class NotRunningError(MachineError):
    """The named experiment is not currently running."""

    run: str
    current: RunStatus

    def __str__(self) -> str:
        return (
            f"{self.run} is not running on {self.host}."
            f" Current status is {self.current}."
        )


@dataclass
class MachineBusyError(MachineError):
    "The machine is busy."
    current: RunStatus

    def __str__(self) -> str:
        return f"Machine {self.host}:{self.port} is currently busy: {self.current}."


@dataclass
class AlreadyExistsError(MachineError):
    "A run already exists in uncollected (experiment:) with the same name."
    name: str

    def __str__(self) -> str:
        return (
            f"Run {self.name} exists in machine working directory. This may be caused by "
            "a failed previous run.  TODO: write cleanup code."
        )


def _find_or_raise(e: ET.ElementTree | ET.Element, path: str) -> ET.Element:
    "Find an element at path and return it, or raise an error."
    x = e.find(path)
    if x is None:
        raise ValueError(f"{path} not found in {x}.")
    else:
        return x


_TABLE_FORMAT = {"markdown": "pipe", "org": "orgtbl"}


class Experiment:
    """A QuantStudio experiment / EDS file

    This class can create, modify, load and save experiments in several ways, run them,
    and control and modify them while running.

    Experiments can be loaded from a file:

    >>> exp: Experiment = Experiment.from_file("experiment.eds")

    They can also be loaded from a running experiment:

    >>> machine = Machine("localhost", 7000, password="password")
    >>> exp = Experiment.from_running(machine)

    Or from the machine's storage:

    >>> exp = Experiment.from_machine(machine, "experiment")

    They can also be created from scratch:

    >>> exp = Experiment("an-experiment-name")
    >>> exp.protocol = Protocol([Stage([Step(time=60, temperature=60)])])
    >>> exp.plate_setup = PlateSetup({"sample_name": "A5"})

    And they can be run on a machine:

    >>> exp.run(machine)

    Data can be accessed in a few ways:

    The (hopefully) easiest way is with welldata, which has multi-indexes
    for both rows and columns.

    >>> exp.welldata.loc[('x1-m4', 4), [('time', 'hours'), ('A05', 'fl')]].plot()

    Or for a temperature curve:

    >>> exp.welldata.loc[('x1-m4', 4), [('A05', 'st'), ('A05', 'fl')]].plot()

    filterdata should still work normally.

    Notes
    -----

    There are a few differences in how QSLib and AB's software handles experiments.

    * AB's software considers the run as starting when the machine indicates "Run Starting".
      This is stored as :any:`runstarttime` in QSLib, but as it may include the lamp warmup
      (3 minutes) and other pre-actual-protocol time, QSLib instead prefers :any:`activestarttime`,
      which it sets from the beginning of the first real (not PRERUN) Stage, at which point
      the machine starts its own active clock and starts ramping to the first temperature.
      QSLib uses this as the start time reference in its data, and also includes the
      timestamp from the machine.
    * The machine has a specific language for run protocols. QSLib uses this language.
      AB's Design and Analysis software *does not*, instead using an XML format. Not
      everything in the machine's language is possible to express in the XML format (eg,
      disabling pcr analysis, saving images); the XML format has some concepts not present in
      the machine format, and is generally more complicated and harder to understand. QSLib uses
      and trusts the machine's protocol if at all possible, *even for files written by
      AB D&A* (it is stored in the log if the run has started).
    * QSLib will try to write a reasonable XML protocol for AB D&A to see, but it may
      by an approximation or simply wrong, if the actual protocol can't be expressed there.
      It will also store the actual protocol in tcprotocol.xml, and its own representation.
    * By default, creating a step with a per-cycle increment in QSLib starts the change
      *on cycle 2*, not cycle 1, as is the default in the software.
    * Immediate pause/resume, mid-run stage addition, and other functions are not supported
      by AB D&A and experiments using them may confuse the software later.
    * QSLib writes notes to XML files, and tries to create reasonable XML files for
      AB D&A, but may still cause problems. At the moment, it makes clear that its files
      are its own (setting software versions in experiment.xml and Manifest.mf).
    """

    protocol: Protocol
    """
    Temperature and reading protocol for the experiment.
    """
    name: str
    """
    Experiment name, also used as file name on the machine.
    """
    runstarttime: datetime | None
    """
    The run start time as a datetime. This is taken *directly from the log*, ignoring the software-set
    value and replacing it on save if possibe.  It is defined as the moment the machine
    records "Run Starting" in its log, using its timestamp.  This may be 3 minutes before
    the start of the protocol if the lamp needs to warm up.  It should be the same value as defined by
    AB's software.

    Use :any:`activestarttime` for a more accurate value.

    None if the file has not been updated since the start of the run
    """
    runendtime: datetime | None
    """
    The run end time as a datetime, taken from the log.  This is the end of the run,
    """
    activestarttime: datetime | None
    """
    The actual beginning of the first stage of the run, defined as the first "Run Stage" message
    in the log after "Stage PRERUN". This is *not* what AB's software considers the start of a run.
    """
    activeendtime: datetime | None
    """
    The actual end of the main part of the run, indicated by "Stage POSTRun" or an abort.
    """
    createdtime: datetime
    """
    The run creation time.
    """
    modifiedtime: datetime
    """
    The last modification time. QSLib sets this on write. AB D&A may not.
    """
    runstate: Literal["INIT", "RUNNING", "COMPLETE", "ABORTED", "STOPPED", "UNKNOWN"]
    """
    Run state, possible values INIT, RUNNING, COMPLETE, ABORTED, STOPPED(?).
    """
    machine: Machine | None
    plate_setup: PlateSetup
    """
    Plate setup for the experiment.
    """
    writesoftware: str
    """
    A string describing the software and version used to write the file.
    """
    _welldata: pd.DataFrame | None = None

    temperatures: pd.DataFrame | None = None
    """
    A DataFrame of temperature readings, at one second resolution, during the experiment
    (and potentially slightly before and after, if included in the message log).

    Columns (as multi-index):

    ("time", ...) : float
        Time of temperature reading, for choices of "timestamp" (Unix timestamp in seconds),
        "seconds" (seconds since the *active* start of the run), or "hours". The latter two may
        be negative, and may not be set if the run never became active.

    ("sample", ...) : float
        Sample temperature for blocks 1, 2, ..., 6, and average in "avg".

    ("block", ...) : float
        Block temperature for blocks 1, 2, ..., 6, and average in "avg".

    ("other", "cover") : float
        Cover temperature

    ("other", "heatsink") : float
        Heatsink temperature

    """

    num_zones: int
    """The number of temperature zones (excluding cover), or -1 if not known."""

    @property
    def all_filters(self) -> Collection[FilterSet]:
        """All filters used at some point in the experiment.

        If the experiment has data, this is based on the existing data.  Otherwise, it is based
        on the experiment protocol.
        """
        if self._welldata is not None:
            return [
                FilterSet.fromstring(f)
                for f in self.welldata.index.get_level_values(0).unique()
            ]
        return self.protocol.all_filters

    @property
    def filter_strings(self) -> list[str]:
        """All filters, as x?-m? strings, used at some point in the experiment."""
        return [str(f) for f in self.all_filters]

    @property
    def welldata(self) -> pd.DataFrame:
        """
        A DataFrame with fluorescence reading information.

        Indices (multi-index) are (filter_set, stage, cycle, step, point), where filter_set
        is a string in familiar form (eg, "x1-m4") and the rest are int.

        Columns (as multi-index):

        ("time", ...) : float
            Time of the data collection, taken from the .quant file.  May differ for different
            filter sets.  Options are "timestamp" (unix timestamp in seconds), "seconds", and
            "hours" (the latter two from the *active* start of the run).

        (well, option) : float
            Data for a well, with well formatted like "A05".  Options are "rt" (read temperature
            from .quant file), "st" (more stable temperature), and "fl" (fluorescence).

        ("exposure", "exposure") : float
            Exposure time from filterdata.xml.  Misleading, because it only refers to the
            longest exposure of multiple exposures.
        """
        if self._welldata is not None:
            return self._welldata
        elif self.runstate == "INIT":
            raise ValueError("Run hasn't started yet: no data available.")
        else:
            raise ValueError("Experiment data is not available")

    def summary(self, format: str = "markdown", plate: str = "list") -> str:
        return self.info(format, plate)

    def info(self, format: str = "markdown", plate: str = "list") -> str:
        """Generate a summary of the experiment, with some formatting configuation. `str()`
        uses this with default parameters.

        Parameters
        ----------
        format : "markdown" or "org", optional
            Format of output, currently "markdown" or "org", and currently matters only
            when plate is "table". By default "markdown".  If an unknown value,
            passed as tablefmt to tabulate.
        plate : "list" or "table", optional
            Format of plate information. "list" gives a list of samples, "table" outputs a
            plate layout table (possibly quite wide). By default "list".

        Returns
        -------
        str
            Summary
        """
        s = f"# QS Experiment {self.name} ({self.runstate})\n\n"
        s += f"{self.protocol}\n\n"
        if plate == "list":
            s += f"{self.plate_setup}\n"
        elif plate == "table":
            s += f"{self.plate_setup.to_table(tablefmt=_TABLE_FORMAT.get(format, format))}\n\n"
        s += f"- Created: {self.createdtime}\n"
        if self.runstarttime:
            s += f"- Run Started: {self.runstarttime}\n"
        if self.runendtime:
            s += f"- Run Ended: {self.runendtime}\n"
        s += f"- Written by: {self.writesoftware}\n"
        s += f"- Read by: QSLib {__version__}\n"
        return s

    def info_html(self) -> str:
        """Create a self-contained HTML summary (returned as a string, but very large) of the experiment."""
        summary = self.info(plate="table")

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(figsize=(21.0 / 2.54, 15.0 / 2.54))
        self.protocol.plot_protocol(ax)

        o = io.StringIO()
        fig.savefig(o, format="svg")
        inner = re.search("<svg.*</svg>", o.getvalue(), re.DOTALL)
        if inner is None:
            raise Exception
        innersvg = inner[0]

        import markdown

        ost = markdown.markdown(summary, output_format="xhtml", extensions=["tables"])

        s = f"""
<!DOCTYPE html>
<html>
<head>
<title>{self.name}</title>
<style>
table {{
    font-size: 10pt;
    border-collapse: collapse;
    margin-left: auto;
    margin-right: auto;
}}
svg {{
    margin-left: auto;
    margin-right: auto;
}}
td:first-of-type {{
    font-weight: bold;
}}
table, th, td {{
    border: 1px solid black;
}}
</style>
</head>
<body>
{ost}

{innersvg}
</body>
</html>
        """

        return s

    def __str__(self) -> str:
        return self.info()

    def _repr_markdown_(self) -> str:
        if len(self.plate_setup.sample_wells) <= 12:
            return self.info()
        else:
            return self.info(plate="table")

    @property
    def runtitle_safe(self) -> str:
        """Run name with " " replaced by "_"."""
        return self.name.replace(" ", "_")

    @property
    def rawdata(self) -> pd.DataFrame:
        warn("rawdata is deprecated; use welldata instead")
        if (self.activestarttime is None) or (self.welldata is None):
            raise DataError
        return _fdc_to_rawdata(
            self.welldata,
            self.activestarttime.timestamp(),
        )

    @property
    def filterdata(self) -> pd.DataFrame:
        warn("filterdata is deprecated; use welldata instead")
        return self.rawdata

    def _ensure_machine(
        self,
        machine: MachineReference | None,
        password: str | None = None,
        needed_level: AccessLevel = AccessLevel.Observer,
    ) -> Machine:
        if isinstance(machine, Machine):
            self.machine = machine
            self.machine.max_access_level = max(
                self.machine.max_access_level, needed_level
            )
            return machine
        elif isinstance(machine, str):
            self.machine = Machine(
                machine,
                password=password,
                max_access_level=needed_level,
            )
            return self.machine
        elif hasattr(self, "machine") and (c := self.machine):
            c.max_access_level = max(c.max_access_level, needed_level)
            return c
        else:
            raise ValueError(
                "No stored machine info for this experiment: please provide some."
            )

    def _ensure_running(self, machine: Machine) -> RunStatus:
        crt = machine.run_status()
        if crt.name != self.runtitle_safe:
            raise NotRunningError(machine, self.runtitle_safe, crt)
        return crt

    def _populate_folder(self, machine: Machine) -> None:
        tempzip = io.BytesIO()
        self.save_file(tempzip)
        exppath = self.runtitle_safe + "/"
        machine.run_command_bytes(
            b"EXP:ZIPWRITE "
            + exppath.encode()
            + b" <message>\n"
            + base64.encodebytes(tempzip.getvalue())
            + b"</message>\n"
        )

    def run(
        self,
        machine: MachineReference | None = None,
        require_exclusive: bool = False,
        require_drawer_check: bool = True,
    ) -> None:
        """Load the run onto a machine, and start it.

        Parameters
        ----------
        machine
            The machine to run on, by default None, in which case the machine
            associated with the run (if any) is used.

        Raises
        ------
        MachineBusyError
            The machine isn't idle.
        AlreadyExistsError
            The machine already has a folder for this run in its working runs folder.
        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        log.info("Attempting to start %s on %s.", self.runtitle_safe, machine.host)

        with machine.ensured_connection():
            # Ensure machine isn't running:
            if (x := machine.run_status()).state.upper() != "IDLE":
                raise MachineBusyError(machine, x)

            if self.runstate != "INIT":
                raise ValueError

            with machine.at_access(AccessLevel.Controller, exclusive=require_exclusive):
                log.debug("Powering on machine and ensuring drawer/cover is closed.")
                # Ensure machine state and power.
                machine.power = True

                machine.drawer_close(lower_cover=True, check=require_drawer_check)

                # Check existence of previous run folder
                if self.runtitle_safe + "/" in machine.run_command("EXP:LIST?"):
                    raise AlreadyExistsError(machine, self.runtitle_safe)

                log.debug("Creating experiment folder %s.", self.runtitle_safe)
                machine.run_command(f"EXP:NEW {self.runtitle_safe} {TEMPLATE_NAME}")

                log.debug("Setting up experiment.xml")
                self.runstarttime = datetime.now()
                self.runstate = "RUNNING"
                self._update_files()

                log.debug("Populating experiment folder.")
                self._populate_folder(machine)

                log.debug("Sending protocol.")
                machine.define_protocol(self.protocol)

                log.debug("Sending run command.")
                # Start the run
                machine.run_command_to_ack(
                    f"RP -samplevolume={self.protocol.volume} -runmode={self.protocol.runmode}"
                    f" {self.protocol.name} {self.runtitle_safe}"
                )  # TODO: user, cover
                log.info("Run %s started on %s.", self.runtitle_safe, machine.host)

    def pause_now(self, machine: MachineReference | None = None) -> None:
        """
        If this experiment is running, pause it (immediately).

        Requires and takes exclusive Controller access on the machine.

        Raises
        ------
        NotRunningError
            the experiment is not currently running
        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        with machine.ensured_connection(AccessLevel.Controller):
            self._ensure_running(machine)
            machine.pause_current_run()

    def resume(self, machine: MachineReference | None = None) -> None:
        """
        If this experiment is running, resume it.

        Requires and takes exclusive Controller access on the machine.

        Raises
        ------
        NotRunningError
            the experiment is not currently running
        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        with machine.ensured_connection(AccessLevel.Controller):
            self._ensure_running(machine)
            machine.resume_current_run()

    def stop(self, machine: MachineReference | None = None) -> None:
        """
        If this experiment is running, stop it after the end of the current cycle.

        Requires and takes exclusive Controller access on the machine.

        Raises
        ------
        NotRunningError
            the experiment is not currently running
        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        with machine.ensured_connection(AccessLevel.Controller):
            self._ensure_running(machine)
            machine.stop_current_run()

    def abort(self, machine: MachineReference | None = None) -> None:
        """
        If this experiment is running, abort it, stopping it immediately.

        Requires and takes exclusive Controller access on the machine.

        Raises
        ------
        NotRunningError
            the experiment is not currently running
        """
        machine = self._ensure_machine(machine)
        with machine.ensured_connection(AccessLevel.Controller):
            self._ensure_running(machine)
            machine.abort_current_run()

    def get_status(self, machine: MachineReference | None = None) -> RunStatus:
        """
        Return the status of the experiment, if currently running.

        Requires Observer access on the machine.

        Raises
        ------
        NotRunningError
            the experiment is not currently running
        """
        machine = self._ensure_machine(machine)
        with machine.ensured_connection(AccessLevel.Observer):
            return self._ensure_running(machine)

    @classmethod
    def _from_running_via_sync(
        cls, machine: MachineReference, include_tiffs: bool = False
    ) -> "Experiment":
        exp = cls(_create_xml=False)

        p = Path(exp._dir_eds)
        p.mkdir(parents=True)

        machine = exp._ensure_machine(machine)

        with machine.ensured_connection():
            crt: str | None = machine.current_run_name

            if not crt:
                raise ValueError("Nothing is currently running.")

            exp.name = crt

            exp.sync_from_machine(
                machine, log_method="copy", include_tiffs=include_tiffs
            )

            exp._update_from_files()

        return exp

    def sync_from_machine(
        self,
        machine: MachineReference | None = None,
        log_method: Literal["copy", "eval"] = "eval",
        include_tiffs: bool = False,
    ) -> None:
        """
        Try to synchronize the data in the experiment to the current state of the run on a
        machine, more efficiently than reloading everything.
        """
        machine = self._ensure_machine(machine)

        with machine.ensured_connection(AccessLevel.Observer):
            # Get a list of all the files in the experiment folder
            machine_files = machine.list_files(
                f"experiments:{self.runtitle_safe}/", verbose=True, recursive=True
            )

            # Transfer anything we don't have
            for f in machine_files:
                name: str = os.path.basename(f["path"])

                if name.lower().endswith("tiff") and not include_tiffs:
                    continue

                sdspath = self._sdspath(
                    re.sub(".*/apldbio/sds/(.*)$", r"\1", f["path"])
                )
                log.debug(f"checking {f['path']} mtime {f['mtime']} to {sdspath}")
                if os.path.exists(sdspath) and os.path.getmtime(sdspath) >= float(
                    f["mtime"]
                ):
                    log.debug(f"{sdspath} has {os.path.getmtime(sdspath)}")
                    continue
                from pathlib import Path  # FIXME

                log.info(f"Updating {sdspath}")

                ldir = f["path"].split("/")[-2]
                if ldir != "sds":
                    cp = Path(self._dir_eds) / ldir
                    if not cp.exists():
                        cp.mkdir()
                if log_method == "copy" or (not f["path"].endswith("messages.log")):
                    with open(sdspath, "wb") as b:
                        b.write(machine.read_file(f["path"]))
                elif log_method == "eval":
                    with open(sdspath, "ab+") as b:

                        # Seeking to the end of the file gives us the size
                        curpos = b.seek(0, 1)

                        # If we're less than 20 bytes, just copy the file:
                        if curpos < 20:
                            b.seek(0)
                            b.write(machine.read_file(f["path"]))
                        else:
                            # Now seek 20 bytes back, and read those 20 bytes
                            b.seek(-20, 1)
                            checklocal = b.read(20)

                            assert b.seek(0, 1) == curpos

                            rdat = machine._get_log_from_byte(
                                self.runtitle_safe, curpos - 20
                            )

                            if rdat[0:20] == checklocal:
                                b.write(rdat[20:])
                            else:
                                log.error(
                                    "Log comparison failure in eval method,"
                                    f" falling back to copy. ('{checklocal.decode()}' ≠ '{rdat[0:20].decode()}')"
                                )
                                b.seek(0)
                                b.write(machine.read_file(f["path"]))
                else:
                    raise NotImplementedError
                os.utime(sdspath, (f["atime"], f["mtime"]))

            # The message log is tricky. Ideally we'd use rsync or wc+tail. TODO
            self._update_from_files()

    def change_protocol_from_now(
        self, new_stages: Sequence[Stage], machine: MachineReference | None = None
    ) -> None:
        """
        For a running experiment, change the remaining stages to be the provided :param:`new_stages` list.
        This is a convenience function that:

        1. Gets the currently-running stage and cycle.
        2. Sets the repeat number of the current stage to its current cycle, thus ending it after the end of
           the current cycle.
        3. Changes the remainder of the stages to be those in the :param:`new_stages` list.

        Because this does not impact any current or past stages, there is less of a need to ensure that the
        stages provided are compatible with the old protocol.  The only check done is to ensure that, if
        the provided stages have any collection commands using default filters, the old protocol has specified
        default filters.  This function does not allow the default filters to be changed: if you want to use
        filters other than the defaults, or defaults were not provided in the old protocol, then either specify
        filters explicitly (recommended) for new stages you'd like to be different, or use
        :any:`Experiment.change_protocol` directly.
        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        with machine.ensured_connection(AccessLevel.Controller):
            self._ensure_running(machine)

            proto = machine.get_running_protocol()

            runstatus = machine.run_status()

            proto.stages[-1].repeat = runstatus.cycle
            proto.stages += new_stages

            machine.define_protocol(proto)
            self.protocol = proto

            self._update_tcprotocol_xml()

            # Push new tcprotocol.xml
            machine.write_file(
                "${LogFolder}/tcprotocol.xml",
                open(self._sdspath("tcprotocol.xml"), "rb").read(),
            )
            machine.write_file(
                "${LogFolder}/qsl-tcprotocol.xml",
                open(self._sdspath("qsl-tcprotocol.xml"), "rb").read(),
            )

    def change_protocol(
        self,
        new_protocol: Protocol,
        machine: MachineReference | None = None,
        force: bool = False,
    ) -> None:
        """
        For a running experiment and an updated protocol, check compatibility
        with the current run, and if possible, update the protocol in the experiment
        file and on the machine, changing the current run.

        Changes that should be possible:

        * Changing the number of cycles of the current run to a higher or lower value (but
          higher or equal to the current cycle number), allowing stages to be lengthened,
          shortened, or stopped.
        * Adding new stages after the current stage.
        * Arbitrarily changing any stage that hasn't started.

        For safest results, ensure power saving is turned off in the Android software.

        Parameters
        ----------
        new_protocol : Protocol
            [description]
        machine : Machine, optional
            [description], by default None

        """
        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
        with machine.ensured_connection(AccessLevel.Controller):
            if not force:
                self._ensure_running(machine)

            runstatus = machine.run_status()

            # Get running protocol and status.
            machine_proto = machine.get_running_protocol()

            # FIXME: can't just do this, may have extra default things added
            # if machine_proto != self.protocol:
            #    raise ValueError("Machine and experiment protocols differ.")

            if not force:
                # Check compatibility of changes.
                machine_proto.check_compatible(new_protocol, runstatus)

            new_protocol.name = machine_proto.name
            new_protocol.volume = machine_proto.volume
            new_protocol.runmode = machine_proto.runmode

            # Make changes.
            machine.define_protocol(new_protocol)

            # Save changes in tcprotocol.xml
            self.protocol = new_protocol

            self._update_tcprotocol_xml()

            # Push new tcprotocol.xml
            machine.write_file(
                "${LogFolder}/tcprotocol.xml",
                open(self._sdspath("tcprotocol.xml"), "rb").read(),
            )
            machine.write_file(
                "${LogFolder}/qsl-tcprotocol.xml",
                open(self._sdspath("qsl-tcprotocol.xml"), "rb").read(),
            )

    def save_file(
        self,
        path_or_stream: str | os.PathLike[str] | IO[bytes] = ".",
        overwrite: bool = False,
        update_files: bool = True,
    ) -> None:
        """
        Save an EDS file of the experiment. This *should* be readable by AB's software,
        but makes no attempt to hide that it was written by QSLib, and contains some other
        information. By default, this will refuse to overwrite an existing file.

        Parameters
        ----------
        path_or_stream : str or os.PathLike[str] or IO[bytes]
            A filename, open binary IO, or directory.  If a directory, the file will
            be saved with the name from `Experiment.runtitle_safe`.
        overwrite : bool, optional
            If True, overwrite any existing file without warning. Defaults to False.
        update_files: bool, optional
            If True (default), update files before saving. Use False if you don't want
            qslib to touch the files, for example, if you have just loaded a run from
            the machine and don't want qslib to change anything based on its interpretation.
        """
        mode: Literal["w", "x"]
        if overwrite:
            mode = "w"
        else:
            mode = "x"

        if update_files:
            self._update_files()

        if isinstance(path_or_stream, IO):
            path: Union[Path, IO] = path_or_stream
        else:
            path = Path(path_or_stream)
            if path.is_dir():
                path = (path / self.runtitle_safe).with_suffix(".eds")

        with zipfile.ZipFile(path, mode) as z:
            for root, _, files in os.walk(self._dir_base):
                for file in files:
                    fpath = os.path.join(root, file)
                    z.write(fpath, os.path.relpath(fpath, self._dir_base))

    # def save_file_without_changes(
    #     self, path: str | os.PathLike[str] | IO[bytes], overwrite: bool = False
    # ) -> None:
    #     """
    #     Save an EDS file of the experiment. Unlike :any:`save_file`, this will not
    #     update any parts of the file, so if it has not been modified elsewhere,
    #     it will be the same as when it was loaded. By default, this will refuse to
    #     overwrite an existing file.

    #     Parameters
    #     ----------
    #     path_or_stream : str or os.PathLike[str] or IO[bytes]
    #         A filename, open binary IO, or directory.  If a directory, the file will
    #         be saved with the name from `Experiment.runtitle_safe`.
    #     overwrite : bool, optional
    #         If True, overwrite any existing file without warning. Defaults to False.
    #     """

    #     return self.save_file(path, overwrite, update_files=False)

    @property
    def sample_wells(self) -> dict[str, list[str]]:
        """A dictionary of sample names to sample wells (convenience read/write access to the :class:`PlateSetup` ."""
        return self.plate_setup.sample_wells

    @sample_wells.setter
    def sample_wells(self, new_sample_wells: dict[str, list[str]]) -> None:
        raise NotImplementedError
        # self.plate_setup.sample_wells = new_sample_wells

    def __init__(
        self,
        name: str | None = None,
        protocol: Protocol | None = None,
        plate_setup: PlateSetup | None = None,
        _create_xml: bool = True,
    ):
        self._tmp_dir_obj = tempfile.TemporaryDirectory()
        self._dir_base = self._tmp_dir_obj.name
        self._dir_eds = os.path.join(self._dir_base, "apldbio", "sds")

        self._protocol_from_qslib: Protocol | None = None
        self._protocol_from_log: Protocol | None = None
        self._protocol_from_xml: Protocol | None = None
        self.runstarttime: datetime | None = None
        self.runendtime: datetime | None = None
        self.activestarttime: datetime | None = None
        self.activeendtime: datetime | None = None

        self.machine: Machine | None = None

        if name is not None:
            self.name = name
        else:
            self.name = _nowuuid()

        if protocol is None:
            self.protocol = Protocol([Stage([Step(60, 25)])])
        else:
            self.protocol = protocol

        if plate_setup:
            self.plate_setup = plate_setup
        else:
            self.plate_setup = PlateSetup()

        if _create_xml:
            self.createdtime = datetime.now()
            self.modifiedtime = self.createdtime
            self.runstarttime = None
            self.runendtime = None
            self.runstate = "INIT"
            self.user: str | None = None
            self.writesoftware = f"QSLib {__version__}"

            self._new_xml_files()
            self._update_files()

    def _new_xml_files(self) -> None:
        os.mkdir(os.path.join(self._dir_base, "apldbio"))
        os.mkdir(os.path.join(self._dir_base, "apldbio", "sds"))

        e = ET.ElementTree(ET.Element("Experiment"))

        ET.SubElement(e.getroot(), "Label").text = "ruo"
        tp = ET.SubElement(e.getroot(), "Type")
        ET.SubElement(tp, "Id").text = "Custom"
        ET.SubElement(tp, "Name").text = "Custom"
        ET.SubElement(tp, "Description").text = "Custom QSLib experiment"
        ET.SubElement(tp, "ResultPersisterName").text = "scAnalysisResultPersister"
        ET.SubElement(
            tp, "ContributedResultPersisterName"
        ).text = "mcAnalysisResultPersister"
        ET.SubElement(e.getroot(), "ChemistryType").text = "Other"
        ET.SubElement(e.getroot(), "TCProtocolMode").text = "Standard"
        ET.SubElement(e.getroot(), "DNATemplateType").text = "WET_DNA"
        ET.SubElement(e.getroot(), "InstrumentTypeId").text = "appletini"  # note cap!
        ET.SubElement(e.getroot(), "BlockTypeID").text = "18"
        ET.SubElement(e.getroot(), "PlateTypeID").text = "TYPE_8X12"

        e.write(open(os.path.join(self._dir_eds, "experiment.xml"), "wb"))

        p = ET.parse(
            io.BytesIO(
                b"""<Plate>
    <Name></Name>
    <BarCode></BarCode>
    <Description></Description>
    <Rows>8</Rows>
    <Columns>12</Columns>
    <PlateKind>
        <Name>96-Well Plate (8x12)</Name>
        <Type>TYPE_8X12</Type>
        <RowCount>8</RowCount>
        <ColumnCount>12</ColumnCount>
    </PlateKind>
    <FeatureMap>
        <Feature>
            <Id>marker-task</Id>
            <Name>marker-task</Name>
        </Feature>
    </FeatureMap>
    </Plate>
    """
            )
        )

        p.write(open(os.path.join(self._dir_eds, "plate_setup.xml"), "wb"))

        ET.ElementTree(ET.Element("TCProtocol")).write(
            open(os.path.join(self._dir_eds, "tcprotocol.xml"), "wb")
        )

        ET.ElementTree(ET.Element("QSLTCProtocol")).write(
            open(os.path.join(self._dir_eds, "qsl-tcprotocol.xml"), "wb")
        )

        with open(os.path.join(self._dir_eds, "Manifest.mf"), "w") as f:
            f.write(_MANIFEST_CONTENTS)

        # File will not load in AB without this, even though it is
        # useless for our purposes.
        with open(self._sdspath("analysis_protocol.xml"), "w") as f:
            f.write(_ANALYSIS_PROTOCOL_TEXT)

    def _sdspath(self, path: str) -> str:
        return os.path.join(self._dir_eds, path)

    def _update_files(self) -> None:
        self._update_experiment_xml()
        self._update_tcprotocol_xml()
        self._update_platesetup_xml()

    def _update_from_files(self) -> None:
        p = Path(self._dir_eds)
        if (p / "experiment.xml").is_file():
            self._update_from_experiment_xml()
        if (p / "tcprotocol.xml").is_file():
            self._update_from_tcprotocol_xml()
        if (p / "plate_setup.xml").is_file():
            self._update_from_platesetup_xml()
        if (p / "messages.log").is_file():
            self._update_from_log()
        self._update_from_data()

        if self._protocol_from_xml:
            self.protocol = self._protocol_from_xml
        if self._protocol_from_log:
            self.protocol = self._protocol_from_log
        if self._protocol_from_qslib:
            self.protocol = self._protocol_from_qslib
        if self._protocol_from_xml:
            self.protocol.covertemperature = self._protocol_from_xml.covertemperature

    @classmethod
    def from_file(cls, file: str | os.PathLike[str] | IO[bytes]) -> Experiment:
        """Load an experiment from an EDS file.

        Returns
        -------
        file : str or os.PathLike[str] or IO[bytes]
            The filename or file handle to read.

        Raises
        ------
        ValueError
            if the file does not appear to be an EDS file (lacks an experiment.xml).
        """
        exp = cls(_create_xml=False)

        with zipfile.ZipFile(file) as z:
            # Ensure that this actually looks like an EDS file:
            try:
                z.getinfo("apldbio/sds/experiment.xml")
            except KeyError:
                raise ValueError(f"{file} does not appear to be an EDS file.") from None

            z.extractall(exp._dir_base)

        exp._update_from_files()

        return exp

    @classmethod
    def from_running(cls, machine: MachineReference) -> "Experiment":
        """Create an experiment from the one currently running on a machine.

        Parameters
        ----------
        machine : Machine
            the machine to connect to

        Returns
        -------
        Experiment
            a copy of the running experiment
        """
        exp = cls(_create_xml=False)

        machine = exp._ensure_machine(machine)

        with machine.ensured_connection():
            crt = machine.current_run_name

            if not crt:
                raise ValueError("Nothing is currently running.")

            z = machine.read_dir_as_zip(crt, leaf="EXP")

            z.extractall(exp._dir_base)

            exp._update_from_files()

        return exp

    @classmethod
    def from_uncollected(
        cls, machine: MachineReference, name: str, move: bool = False
    ) -> Experiment:
        """Create an experiment from the uncollected (not yet compressed)
        storage.

        Parameters
        ----------
        machine : Machine
            the machine to connect to

        name: str
            the name of the run to collect.

        Returns
        -------
        Experiment
            a copy of the experiment
        """
        exp = cls(_create_xml=False)

        machine = exp._ensure_machine(machine)

        with machine.ensured_connection():
            crt = name

            if move:
                raise NotImplementedError

            if not crt:
                raise ValueError("Nothing is currently running.")

            z = machine.read_dir_as_zip(crt, leaf="EXP")

            z.extractall(exp._dir_base)

            exp._update_from_files()

        return exp

    @classmethod
    def from_machine_storage(cls, machine: MachineReference, name: str) -> Experiment:
        """Create an experiment from the one currently running on a machine.

        Parameters
        ----------
        machine : Machine
            the machine to connect to

        Returns
        -------
        Experiment
            a copy of the experiment
        """
        exp = cls(_create_xml=False)

        machine = exp._ensure_machine(machine)

        with machine.ensured_connection():
            try:
                o = machine.read_file(name + ".eds", context="public_run_complete")
            except FileNotFoundError:
                o = machine.read_file(name, context="public_run_complete")

            z = zipfile.ZipFile(io.BytesIO(o))

            z.extractall(exp._dir_base)

            exp._update_from_files()

        return exp

    @classmethod
    def from_machine(cls, machine: MachineReference, name: str) -> Experiment:
        """Create an experiment from data on a machine, checking the running
        experiment if any, the machine's public_run_complete storage, and the
        machine's uncollected storage.

        Parameters
        ----------
        machine : Machine | str
            the machine to connect to, either as a Machine or a host name

        name: str
            the name of the run

        Returns
        -------
        Experiment
            a copy of the experiment
        """
        if isinstance(machine, str):
            machine = Machine(machine)

        with machine.ensured_connection():
            if name == machine.current_run_name:
                exp = cls.from_running(machine)
            elif name in machine.list_runs_in_storage():
                exp = cls.from_machine_storage(machine, name)
            elif name + "/" in machine.list_files("", verbose=False, leaf="EXP"):
                exp = cls.from_uncollected(machine, name)
            else:
                raise FileNotFoundError(f"Could not find run {name} on {machine.host}.")
        return exp

    def _update_experiment_xml(self) -> None:
        exml = ET.parse(os.path.join(self._dir_eds, "experiment.xml"))

        _set_or_create(exml, "Name", text=self.name)
        if self.user:
            _set_or_create(exml, "Operator", text=self.user)
        else:
            if (e := exml.find("Operator")) is not None:
                exml.getroot().remove(e)
        _set_or_create(
            exml, "CreatedTime", str(int(self.createdtime.timestamp() * 1000))
        )
        _set_or_create(
            exml, "ModifiedTime", str(int(datetime.now().timestamp() * 1000))
        )
        if self.runstarttime:
            _set_or_create(
                exml, "RunStartTime", str(int(self.runstarttime.timestamp() * 1000))
            )
        if self.runendtime:
            _set_or_create(
                exml, "RunEndTime", str(int(self.runendtime.timestamp() * 1000))
            )

        _set_or_create(exml, "RunState", self.runstate)

        sinfo = exml.find(
            "ExperimentProperty[@type='RunInfo']/PropertyValue[@key='softwareVersion']/String"
        )
        if sinfo is None:
            if not (e := exml.find("ExperimentProperty[@type='RunInfo']")):
                e = ET.SubElement(exml.getroot(), "ExperimentProperty", type="RunInfo")
            e.append(sinfo := ET.Element("PropertyValue", key="softwareVersion"))
            sinfo = ET.SubElement(sinfo, "String")
        sinfo.text = f"QSLib {__version__}"

        ET.indent(exml)

        exml.write(os.path.join(self._dir_eds, "experiment.xml"))

    def _update_from_experiment_xml(self) -> None:
        exml = ET.parse(os.path.join(self._dir_eds, "experiment.xml"))

        self.name = exml.findtext("Name") or "unknown"
        self.user = exml.findtext("Operator") or None
        self.createdtime = datetime.fromtimestamp(
            float(_find_or_raise(exml, "CreatedTime").text) / 1000.0  # type: ignore
        )
        self.runstate = exml.findtext("RunState") or "UNKNOWN"  # type: ignore
        self.writesoftware = (
            exml.findtext(
                "ExperimentProperty[@type='RunInfo']/PropertyValue[@key='softwareVersion']/String"
            )
            or "UNKNOWN"
        )
        if x := exml.findtext("RunStartTime"):
            self.runstarttime = datetime.fromtimestamp(float(x) / 1000.0)
        if x := exml.findtext("RunEndTime"):
            self.runendtime = datetime.fromtimestamp(float(x) / 1000.0)

    def _update_tcprotocol_xml(self) -> None:
        if self.protocol:
            # exml = ET.parse(os.path.join(self._dir_eds, "tcprotocol.xml"))
            tcxml, qstcxml = self.protocol.to_xml()
            ET.indent(tcxml)
            ET.indent(qstcxml)
            tcxml.write(os.path.join(self._dir_eds, "tcprotocol.xml"))

            # Make new machine with stripped password:
            if self.machine:
                m2d = self.machine.asdict(password=False)
                ET.SubElement(qstcxml.getroot(), "MachineConnection").text = toml.dumps(
                    m2d
                )

            qstcxml.write(os.path.join(self._dir_eds, "qsl-tcprotocol.xml"))

    def _update_from_tcprotocol_xml(self) -> None:
        exml = ET.parse(os.path.join(self._dir_eds, "tcprotocol.xml"))
        if os.path.isfile(os.path.join(self._dir_eds, "qsl-tcprotocol.xml")):
            qstcxml = ET.parse(os.path.join(self._dir_eds, "qsl-tcprotocol.xml"))

            if ((x := qstcxml.find("QSLibProtocolCommand")) is not None) and (
                x.text is not None
            ):
                try:
                    self._protocol_from_qslib = Protocol.from_scpicommand(
                        SCPICommand.from_string(x.text)
                    )
                except ValueError:
                    self._protocol_from_qslib = None
            else:
                self._protocol_from_qslib = None

            if (mc := qstcxml.findtext("MachineConnection")) and not self.machine:
                try:
                    self.machine = Machine(**toml.loads(mc))
                except ValueError:
                    pass
        try:
            self._protocol_from_xml = Protocol.from_xml(exml.getroot())
        except Exception as e:
            print(e)
            self._protocol_from_xml = None

    def _update_platesetup_xml(self) -> None:
        x = ET.parse(os.path.join(self._dir_eds, "plate_setup.xml"))
        self.plate_setup.update_xml(x.getroot())
        ET.indent(x)
        x.write(os.path.join(self._dir_eds, "plate_setup.xml"))

    def _update_from_platesetup_xml(self) -> None:
        x = ET.parse(os.path.join(self._dir_eds, "plate_setup.xml")).getroot()
        self.plate_setup = PlateSetup.from_platesetup_xml(x)

    def _update_from_data(self) -> None:
        fdp = os.path.join(self._dir_eds, "filterdata.xml")
        if os.path.isfile(fdp):
            fdx = ET.parse(fdp)
            fdrs = [
                FilterDataReading(x, sds_dir=self._dir_eds)
                for x in fdx.findall(".//PlateData")
            ]
            self._welldata = df_from_readings(
                fdrs, self.activestarttime.timestamp() if self.activestarttime else None
            )
        elif fdfs := glob(os.path.join(self._dir_eds, "filter", "*_filterdata.xml")):
            fdrs = [
                FilterDataReading.from_file(fdf, sds_dir=self._dir_eds) for fdf in fdfs
            ]
            self._welldata = df_from_readings(
                fdrs, self.activestarttime.timestamp() if self.activestarttime else None
            )
        else:
            self._welldata = None

    def data_for_sample(self, sample: str) -> pd.DataFrame:
        """Convenience function to return data for a specific sample.

        Finds wells using :code:`self.plate_setup.sample_wells[sample]`, then
        returns :code:`self.welldata.loc[:, wells]`

        Parameters
        ----------
        sample : str
            sample name

        Returns
        -------
        pd.Dataframe
            Slice of welldata.  Will have multiple wells if sample is in multiple wells.
        """
        wells = ["time"] + [
            f"{x[0]}{int(x[1:])}" for x in self.plate_setup.sample_wells[sample]
        ]
        x = self.welldata.loc[:, wells]
        return x

    def _update_from_log(self) -> None:
        if not os.path.isfile(os.path.join(self._dir_eds, "messages.log")):
            return
        try:
            msglog = open(os.path.join(self._dir_eds, "messages.log"), "r").read()
        except UnicodeDecodeError as error:
            log.debug(
                "Decoding log failed. If <binary.reply> is present in log this may be the cause:"
                "if so contact Constantine (<const@costinet.org>).  Continuing with failed characters replaced with backslash notation"
                "Failure was in this area:\n"
                "{!r}".format(error.object[error.start - 500 : error.end + 500])
            )
            msglog = open(
                os.path.join(self._dir_eds, "messages.log"),
                "r",
                errors="backslashreplace",
            ).read()

        ms = re.finditer(
            r"^Run (?P<ts>[\d.]+) (?P<msg>\w+)(?: (?P<ext>\S+))?", msglog, re.MULTILINE
        )
        self.runstarttime = None
        self.runendtime = None
        self.prerunstart = None
        self.activestarttime = None
        self.activeendtime = None
        self.runstate = "INIT"

        m: Optional[re.Match[str]]
        stages: list[dict[str, int | str | datetime]] = []

        for m in ms:
            ts = datetime.fromtimestamp(float(m["ts"]))
            if m["msg"] == "Starting":
                self.runstarttime = ts
            elif m["msg"] == "Stage":
                if m["ext"] == "PRERUN":
                    self.prerunstart = ts
                    self.runstate = "RUNNING"
                elif (self.activestarttime is None) and (self.prerunstart is not None):
                    self.activestarttime = ts
                elif m["ext"] == "POSTRun":
                    self.activeendtime = ts
                try:
                    stages.append({"stage": int(m["ext"]), "start_time": ts})
                except ValueError:
                    stages.append({"stage": m["ext"], "start_time": ts})
                if len(stages) > 1:
                    stages[-2]["end_time"] = ts
            elif m["msg"] == "Ended":
                self.runendtime = ts
                self.runstate = "COMPLETE"
                if len(stages) > 1:
                    stages[-1].setdefault("end_time", ts)
            elif m["msg"] == "Aborted":
                self.runstate = "ABORTED"
                self.activeendtime = ts
                self.runendtime = ts
                if len(stages) > 1:
                    stages[-1].setdefault("end_time", ts)
            elif m["msg"] == "Stopped":
                self.runstate = "STOPPED"
                self.activeendtime = ts
                self.runendtime = ts
                if len(stages) > 1:
                    stages[-1].setdefault("end_time", ts)

        self.stages = pd.DataFrame(stages, columns=["stage", "start_time", "end_time"])

        if self.activestarttime:
            self.stages["start_seconds"] = (
                self.stages["start_time"] - self.activestarttime
            ).astype("timedelta64[s]")
            self.stages["end_seconds"] = (
                self.stages["end_time"] - self.activestarttime
            ).astype("timedelta64[s]")

        tt = []

        try:
            # In normal runs, the protocol is there without the PROT command at the
            # beginning of the log as an info command, with quote.message.  Let's
            # try to grab it!
            if m := re.match(
                r"^Info (?:[\d.]+) (<quote.message>.*?</quote.message>)",
                msglog,
                re.DOTALL | re.MULTILINE,
            ):
                # We can get the prot name too, and sample volume! FIXME: not from qslib runs!
                rp = re.search(
                    r"NEXT RP (?:-CoverTemperature=(?P<ct>[\d.]+) )?(?:-SampleVolume=(?P<sv>[\d.]+) )?([\w-]+) (?P<protoname>[\w-]+)",
                    msglog,
                    re.IGNORECASE,
                )
                if rp:
                    prot = Protocol.from_scpicommand(
                        SCPICommand.from_string(f"PROT {rp['protoname']} {m[1]}")
                    )
                    if rp[1]:
                        prot.volume = float(rp["sv"])
                else:
                    prot = Protocol.from_scpicommand(
                        SCPICommand.from_string(f"PROT unknown_name {m[1]}")
                    )
                self._protocol_from_log = prot

                # Now that we know the protocol name, we can search for whether the protocol was changed later:
                for mm in re.finditer(
                    r"^Debug.*<quote.message>C:.* (PROT[ O].*\n(c:.*\n)+.*)</quote.message>",
                    msglog,
                    re.MULTILINE,
                ):
                    newprot = Protocol.from_scpicommand(
                        SCPICommand.from_string(mm[1].replace("\nc:", "\n"))
                    )
                    # if newprot.name == prot.name:
                    self._protocol_from_log = newprot

            else:
                self._protocol_from_log = None
        except ValueError:
            self._protocol_from_log = None

        for m in re.finditer(
            r"^Temperature ([\d.]+) -sample=([\d.,]+) -heatsink=([\d.,]+) "
            r"-cover=([\d.,]+) -block=([\d.,]+)$",
            msglog,
            re.MULTILINE,
        ):
            r = []
            r.append(float(m[1]))
            r += [float(x) for x in m[2].split(",")]
            r.append(float(m[3]))
            r.append(float(m[4]))
            r += [float(x) for x in m[5].split(",")]
            tt.append(r)

        if len(tt) > 0:
            self.num_zones = int((len(tt[0]) - 3) / 2)

            self.temperatures = pd.DataFrame(
                tt,
                columns=pd.MultiIndex.from_tuples(
                    [(cast(str, "time"), cast(Union[str, int], "timestamp"))]
                    + [("sample", n) for n in range(1, self.num_zones + 1)]
                    + [("other", "heatsink"), ("other", "cover")]
                    + [("block", n) for n in range(1, self.num_zones + 1)]
                ),
            )

            if self.activestarttime:
                self.temperatures[("time", "seconds")] = (
                    self.temperatures[("time", "timestamp")]
                    - self.activestarttime.timestamp()
                )
                self.temperatures[("time", "hours")] = (
                    self.temperatures[("time", "seconds")] / 3600.0
                )
            self.temperatures[("sample", "avg")] = self.temperatures["sample"].mean(
                axis=1
            )
            self.temperatures[("block", "avg")] = self.temperatures["block"].mean(
                axis=1
            )

        else:
            self.num_zones = -1
            self.temperatures = None

    def _find_anneal_stages(self) -> list[int]:
        anneal_stages = []
        for i, stage in enumerate(self.protocol.stages):
            if any(
                getattr(step, "collect", False)
                and (getattr(step, "temp_increment", 0.0) < 0.0)
                for step in stage.steps
            ):
                anneal_stages.append(i + 1)
        return anneal_stages

    def _find_melt_stages(
        self, anneal_stages: Sequence[int] | None
    ) -> tuple[list[int], list[int]]:
        melt_stages = []
        for i, stage in enumerate(self.protocol.stages):
            if any(
                getattr(step, "collect", False)
                and (getattr(step, "temp_increment", 0.0) > 0.0)
                for step in stage.steps
            ):
                melt_stages.append(i + 1)

        if (
            (anneal_stages is not None)
            and (len(anneal_stages) > 0)
            and (len(melt_stages) > 0)
            and (max(anneal_stages) + 1 < min(melt_stages))
        ):
            between_stages = [
                x for x in range(max(anneal_stages) + 1, min(melt_stages))
            ]
        else:
            between_stages = []

        return melt_stages, between_stages

    def plot_anneal_melt(
        self,
        samples: str | Sequence[str] | None = None,
        filters: str | FilterSet | Collection[str | FilterSet] | None = None,
        anneal_stages: int | Sequence[int] | None = None,
        melt_stages: int | Sequence[int] | None = None,
        between_stages: int | Sequence[int] | None = None,
        process: Sequence[Processor] | Processor | None = None,
        normalization: Processor | None = None,
        ax: "plt.Axes" | None = None,
        marker: str | None = None,
        legend: bool | Literal["inset", "right"] = True,
        figure_kw: Mapping[str, Any] | None = None,
        line_kw: Mapping[str, Any] | None = None,
    ) -> "plt.Axes":
        """
        Plots anneal/melt curves.

        This uses solid lines for the anneal, dashed lines for the melt, and
        dotted lines for anything "between" the anneal and melt (for example, a temperature
        hold).

        Line labels are intended to provide full information when read in combination with
        the axes title.  They will only include information that does not apply to all
        lines.  For example, if every line is from the same filter set, but different
        samples, then only the sample will be shown.  If every line is from the same sample,
        but different filter sets, then only the filter set will be shown. Wells are shown
        if a sample has multiple wells.

        Parameters
        ----------

        samples
            A reference to a single sample (a string), a list of sample names, or a Python
            regular expression as a string, matching sample names (full start-to-end matches
            only). Well names may also be included, in which case each well will be treated
            without regard to the sample name that may refer to it.  Note this means you cannot
            give your samples names that correspond with well references.  If not provided,
            all (named) samples will be included.

        filters
            Optional. A filterset (string or `FilterSet`) or list of filtersets to include in the
            plot.  Multiple filtersets will be plotted *on the same axes*.  Optional;
            if None, then all filtersets with data in the experiment will be included.

        anneal_stages, melt_stages, between_stages: int | Sequence[int] | None
            Optional. A stage or list of stages (integers, starting from 1), corresponding to the
            anneal, melt, and stages between the anneal and melt (if any).  Any of these
            may be None, in which case the function will try to determine the correct values
            automatically.

        normalization
            Optional. A Normalizer instance to apply to the data.  By default, this is NormRaw, which
            passes through raw fluorescence values.  NormToMeanPerWell also works well.

        ax
            Optional.  An axes to put the plot on.  If not provided, the function will
            create a new figure, by default with constrained_layout=True, though this
            can be modified with figure_kw.

        marker
            The marker format for data points, or None for no markers (default).

        legend
            Whether to add a legend.  True (default) decides whether to have the legend
            as an inset or to the right of the axes based on the number of lines.  "inset"
            and "right" specify the positioning.  Note that for "right", you must use
            some method to adjust the axes positioning: constrained_layout, tight_layout,
            or manually reducing the axes width are all options.

        figure_kw
            Optional.  A dictionary of options passed through as keyword options to
            the figure creation.  Only applies if ax is None.

        line_kw
            Optional.  A dictionary of keywords passed to all three plotting commands.

        Returns
        -------

        plt.Axes
            The axes object of the plot.

        """

        import matplotlib.pyplot as plt

        if process is None:
            process = [normalization or NormRaw()]
        elif normalization:
            raise ValueError(
                "Can't specify both process and normalization (include normalization in process list)."
            )
        if isinstance(process, Processor):
            process = [process]

        if filters is None:
            filters = self.all_filters

        samples = self._get_samples(samples)

        filters = _normalize_filters(filters)

        if anneal_stages is None:
            anneal_stages = self._find_anneal_stages()
        elif isinstance(anneal_stages, int):
            anneal_stages = [anneal_stages]
        else:
            anneal_stages = list(anneal_stages)

        if melt_stages is None:
            melt_stages, between_stages = self._find_melt_stages(anneal_stages)
        elif isinstance(melt_stages, int):
            melt_stages = [melt_stages]
        else:
            melt_stages = list(melt_stages)

        if between_stages is None:
            between_stages = []
        elif isinstance(between_stages, int):
            between_stages = [between_stages]
        else:
            between_stages = list(between_stages)

        if ax is None:
            ax = cast(
                plt.Axes,
                plt.figure(
                    **(
                        {"constrained_layout": True}
                        | (({} if figure_kw is None else figure_kw))
                    )
                ).add_subplot(),
            )

        data = self.welldata

        for processor in process:
            data = processor.process_scoped(data, "all")

        all_wells = self.plate_setup.get_wells(samples)

        reduceddata = data.loc[[f.lowerform for f in filters], all_wells]

        for processor in process:
            reduceddata = processor.process_scoped(reduceddata, "limited")

        for filter in filters:
            filterdat: pd.DataFrame = reduceddata.loc[filter.lowerform, :]  # type: ignore

            annealdat: pd.DataFrame = filterdat.loc[anneal_stages, :]  # type: ignore
            meltdat: pd.DataFrame = filterdat.loc[melt_stages, :]  # type: ignore

            if len(between_stages) > 0:
                betweendat: pd.DataFrame = filterdat.loc[between_stages, :]  # type: ignore

            anneallines = []
            meltlines = []
            betweenlines = []

            for sample in samples:
                wells = self.plate_setup.get_wells(sample)

                for well in wells:
                    color = next(ax._get_lines.prop_cycler)["color"]

                    label = _gen_label(
                        self.plate_setup.get_descriptive_string(sample),
                        well,
                        filter,
                        samples,
                        wells,
                        filters,
                    )

                    anneallines.append(
                        ax.plot(
                            annealdat.loc[:, (well, "st")],
                            annealdat.loc[:, (well, "fl")],
                            color=color,
                            label=label,
                            marker=marker,
                            **(line_kw if line_kw is not None else {}),
                        )
                    )

                    meltlines.append(
                        ax.plot(
                            meltdat.loc[:, (well, "st")],
                            meltdat.loc[:, (well, "fl")],
                            color=color,
                            linestyle="dashed",
                            marker=marker,
                            **(line_kw if line_kw is not None else {}),
                        )
                    )

                    if len(between_stages) > 0:
                        betweenlines.append(
                            ax.plot(
                                betweendat.loc[:, (well, "st")],
                                betweendat.loc[:, (well, "fl")],
                                color=color,
                                linestyle="dotted",
                                marker=marker,
                                **(line_kw if line_kw is not None else {}),
                            )
                        )

        ax.set_xlabel("temperature (°C)")

        ylabel: str | None = None
        for processor in process:
            ylabel = processor.ylabel(ylabel)
        ax.set_ylabel(ylabel or "fluorescence")

        if legend is True:
            if len(anneallines) < 6:
                legend = "inset"
            else:
                legend = "right"

        if legend == "inset":
            ax.legend()
        elif legend == "right":
            ax.legend(bbox_to_anchor=(1.04, 1), loc="upper left")

        ax.set_title(
            _gen_axtitle(
                self.name,
                anneal_stages + between_stages + melt_stages,
                samples,
                all_wells,
                filters,
            )
        )

        return ax

    def _get_samples(self, samples: str | Sequence[str] | None) -> Sequence[str]:
        if isinstance(samples, str):
            if samples in self.plate_setup.sample_wells:
                samples = [samples]
            else:
                samples = [
                    k
                    for k in self.plate_setup.sample_wells
                    if re.match(samples + "$", k)
                ]
                if not samples:
                    raise ValueError(f"Samples not found")
        elif samples is None:
            samples = list(self.plate_setup.sample_wells.keys())
        return samples

    def plot_over_time(
        self,
        samples: str | Sequence[str] | None = None,
        filters: str | FilterSet | Collection[str | FilterSet] | None = None,
        stages: slice | int | Sequence[int] = slice(None),
        process: Sequence[Processor] | Processor | None = None,
        normalization: Processor | None = None,
        ax: "plt.Axes" | "Sequence[plt.Axes]" | None = None,
        legend: bool | Literal["inset", "right"] = True,
        temperatures: Literal[False, "axes", "inset", "twin"] = "axes",
        marker: str | None = None,
        stage_lines: bool | Literal["fluorescence", "temperature"] = True,
        annotate_stage_lines: (
            bool
            | float
            | Literal["fluorescence", "temperature"]
            | Tuple[Literal["fluorescence", "temperature"], float]
        ) = False,
        figure_kw: Mapping[str, Any] | None = None,
        line_kw: Mapping[str, Any] | None = None,
    ) -> "Sequence[plt.Axes]":
        """
        Plots fluorescence over time, optionally with temperatures over time.

        Line labels are intended to provide full information when read in combination with
        the axes title.  They will only include information that does not apply to all
        lines.  For example, if every line is from the same filter set, but different
        samples, then only the sample will be shown.  If every line is from the same sample,
        but different filter sets, then only the filter set will be shown. Wells are shown
        if a sample has multiple wells.

        Parameters
        ----------

        samples
            A reference to a single sample (a string), a list of sample names, or a Python
            regular expression as a string, matching sample names (full start-to-end matches
            only). Well names may also be included, in which case each well will be treated
            without regard to the sample name that may refer to it.  Note this means you cannot
            give your samples names that correspond with well references.  If not provided,
            all (named) samples will be included.

        filters
            Optional. A filterset (string or `FilterSet`) or list of filtersets to include in the
            plot.  Multiple filtersets will be plotted *on the same axes*.  Optional;
            if None, then all filtersets with data in the experiment will be included.

        stages
            Optional.  A stage, list of stages, or slice (all using integers starting
            from 1), to include in the plot.  By default, all stages are plotted.
            For example, to plot stage 2, use `stages=2`; to plot stages 2 and 4, use
            `stages=[2, 4]`, to plot stages 3 through 15, use `stages=slice(3, 16)`
            (Python ranges are exclusive on the end).  Note that is a slice, you
            can use `None` instead of a number to denote the beginning/end.

        normalization
            Optional. A Normalizer instance to apply to the data.  By default, this is NormRaw, which
            passes through raw fluorescence values.  NormToMeanPerWell also works well.

        temperatures
            Optional (default "axes").  Several alternatives for displaying temperatures.
            "axes" uses a separate axes (created if ax is not provided, otherwise ax must
            be a list of two axes).

            Temperatures are from Experiment.temperature, and are thus the temperatures
            as recorded during the run, not the set temperatures.  Note that this has a
            *very* large number of data points, something that should be dealt with at
            some point.

        ax
            Optional.  An axes to put the plot on.  If not provided, the function will
            create a new figure, by default with constrained_layout=True, though this
            can be modified with figure_kw.  If `temperatures="axes"`, you must provide
            a list or tuple of *two* axes, the first for fluorescence, the second
            for temperature.

        marker
            The marker format for data points, or None for no markers (default).

        legend
            Whether to add a legend.  True (default) decides whether to have the legend
            as an inset or to the right of the axes based on the number of lines.  "inset"
            and "right" specify the positioning.  Note that for "right", you must use
            some method to adjust the axes positioning: constrained_layout, tight_layout,
            or manually reducing the axes width are all options.

        stage_lines
            Whether to include dotted vertical lines on transitions between stages.  If
            "fluorescence" or "temperature", include only on one of the two axes.

        annotate_stage_lines
            Whether to include text annotations for stage lines.  Float parameter allows
            setting the minimum duration of stage, as a fraction of total plotted time, to
            annotate, in order to avoid overlapping annotations (default threshold is 0.05).

        figure_kw
            Optional.  A dictionary of options passed through as keyword options to
            the figure creation.  Only applies if ax is None.

        line_kw
            Optional.  A dictionary of keywords passed to fluorescence plot commands.

        """

        import matplotlib.pyplot as plt

        if process is None:
            process = [normalization or NormRaw()]
        elif normalization:
            raise ValueError(
                "Can't specify both process and normalization (include normalization in process list)."
            )
        if isinstance(process, Processor):
            process = [process]

        if filters is None:
            filters = self.all_filters

        filters = _normalize_filters(filters)

        samples = self._get_samples(samples)

        if isinstance(stages, int):
            stages = [stages]

        fig = None

        if ax is None:
            if temperatures == "axes":
                fig, ax = plt.subplots(
                    2,
                    1,
                    sharex="all",
                    gridspec_kw={"height_ratios": [3, 1]},
                    **(
                        {"constrained_layout": True}
                        | (({} if figure_kw is None else figure_kw))
                    ),
                )
            else:
                fig, ax = plt.subplots(1, 1, **({} if figure_kw is None else figure_kw))
                ax = [ax]

        elif (not isinstance(ax, (Sequence, np.ndarray))) or isinstance(ax, plt.Axes):
            ax = [ax]

        ax = cast(Sequence[plt.Axes], ax)

        data = self.welldata

        for processor in process:
            data = processor.process_scoped(data, "all")

        all_wells = self.plate_setup.get_wells(samples) + ["time"]

        reduceddata = data.loc[[f.lowerform for f in filters], all_wells]

        for processor in process:
            reduceddata = processor.process_scoped(reduceddata, "limited")

        lines = []
        for filter in filters:
            filterdat: pd.DataFrame = reduceddata.loc[filter.lowerform, :]  # type: ignore

            for sample in samples:
                wells = self.plate_setup.get_wells(sample)

                for well in wells:
                    color = next(ax[0]._get_lines.prop_cycler)["color"]

                    label = _gen_label(
                        self.plate_setup.get_descriptive_string(sample),
                        well,
                        filter,
                        samples,
                        wells,
                        filters,
                    )

                    lines.append(
                        ax[0].plot(
                            filterdat.loc[stages, ("time", "hours")],
                            filterdat.loc[stages, (well, "fl")],
                            color=color,
                            label=label,
                            marker=marker,
                            **(line_kw if line_kw is not None else {}),
                        )
                    )

        ax[-1].set_xlabel("time (hours)")

        ylabel: str | None = None
        for processor in process:
            ylabel = processor.ylabel(ylabel)
        ax[0].set_ylabel(ylabel or "fluorescence")

        if legend is True:
            if len(lines) < 6:
                legend = "inset"
            else:
                legend = "right"

        if legend == "inset":
            ax[0].legend()
        elif legend == "right":
            ax[0].legend(bbox_to_anchor=(1.04, 1), loc="upper left")

        xlims = ax[0].get_xlim()

        if isinstance(annotate_stage_lines, (tuple, list)):
            if annotate_stage_lines[0] == "fluorescence":
                fl_asl: bool | float = annotate_stage_lines[1]
                t_asl: bool | float = False
            elif annotate_stage_lines[0] == "temperature":
                fl_asl = False
                t_asl = annotate_stage_lines[1]
            else:
                raise ValueError
        elif annotate_stage_lines == "temperature":
            t_asl = True
            fl_asl = False
        elif annotate_stage_lines == "fluorescence":
            t_asl = False
            fl_asl = True
        else:
            t_asl = annotate_stage_lines
            fl_asl = annotate_stage_lines

        if stage_lines == "temperature":
            t_sl = True
            fl_sl = False
        elif stage_lines == "fluorescence":
            t_sl = False
            fl_sl = True
        else:
            t_sl = stage_lines
            fl_sl = stage_lines

        self._annotate_stages(ax[0], fl_sl, fl_asl, (xlims[1] - xlims[0]) * 3600.0)

        if temperatures == "axes":
            if len(ax) < 2:
                raise ValueError("Temperature axes requires at least two axes in ax")

            xlims = ax[0].get_xlim()
            tmin, tmax = np.inf, 0.0

            for i, fr in reduceddata.groupby("filter_set", as_index=False):
                d = fr.loc[i, ("time", "hours")].loc[stages]
                tmin = min(tmin, d.min())
                tmax = max(tmax, d.max())

            self.plot_temperatures(
                hours=(tmin, tmax),
                ax=ax[1],
                stage_lines=t_sl,
                annotate_stage_lines=t_asl,
            )

            ax[0].set_xlim(xlims)
            ax[1].set_xlim(xlims)
        elif temperatures is False:
            pass
        else:
            raise NotImplementedError

        ax[0].set_title(_gen_axtitle(self.name, stages, samples, all_wells, filters))

        return ax

    def plot_protocol(
        self, ax: Optional[plt.Axes] = None
    ) -> Tuple[plt.Axes, Tuple[List[plt.Line2D], List[plt.Line2D]]]:
        """A plot of the temperature and data collection points in the experiment's protocol."""

        return self.protocol.plot_protocol(ax)

    def plot_temperatures(
        self,
        *,
        sel: slice | Callable[[pd.DataFrame], bool] = slice(None),
        hours: tuple[float, float] | None = None,
        ax: Optional[plt.Axes] = None,
        stage_lines: bool = True,
        annotate_stage_lines: bool | float = True,
        legend: bool = False,
        figure_kw: Mapping[str, Any] | None = None,
        line_kw: Mapping[str, Any] | None = None,
    ) -> "plt.Axes":
        """Plot sample temperature readings.

        Parameters
        ----------

        sel
            A selector for the temperature DataFrame.  This is not necessarily
            easy to use; `hours` is an easier alternative.

        hours
            Constructs a selector to show temperatures for a time range.
            :param:`sel` should not be set.

        ax
            Optional.  An axes to put the plot on.  If not provided, the function will
            create a new figure, by default with constrained_layout=True, though this
            can be modified with figure_kw.

        stage_lines
            Whether to include dotted vertical lines on transitions between stages.

        annotate_stage_lines
            Whether to include text annotations for stage lines.  Float parameter allows
            setting the minimum duration of stage, as a fraction of total plotted time, to
            annotate, in order to avoid overlapping annotations (default threshold is 0.05).

        legend
            Whether to add a legend.

        figure_kw
            Optional.  A dictionary of options passed through as keyword options to
            the figure creation.  Only applies if ax is None.

        line_kw
            Optional.  A dictionary of keywords passed to plot commands.
        """

        import matplotlib.pyplot as plt

        if not hasattr(self, "temperatures") or self.temperatures is None:
            raise ValueError("Experiment has no temperature data.")

        if hours is not None:
            if sel != slice(None):
                raise ValueError("sel and hours cannot both be set.")
            tmin, tmax = hours
            sel = lambda x: (tmin <= x[("time", "hours")]) & (
                x[("time", "hours")] <= tmax
            )

        if ax is None:
            _, ax = plt.subplots(**(figure_kw or {}))

        reltemps = self.temperatures.loc[sel, :]

        for x in range(1, self.num_zones):
            ax.plot(
                reltemps.loc[:, ("time", "hours")],
                reltemps.loc[:, ("sample", x)],
                label=f"zone {x}",
                **(line_kw or {}),
            )

        v = reltemps.loc[:, ("time", "hours")]
        totseconds = 3600.0 * (v.iloc[-1] - v.iloc[0])

        self._annotate_stages(ax, stage_lines, annotate_stage_lines, totseconds)

        ax.set_ylabel("temperature (°C)")
        ax.set_xlabel("time (hours)")

        if legend:
            ax.legend()

        return ax

    def _annotate_stages(self, ax, stage_lines, annotate_stage_lines, totseconds):
        if stage_lines:
            if isinstance(annotate_stage_lines, float):
                annotate_frac = annotate_stage_lines
                annotate_stage_lines = True
            else:
                annotate_frac = 0.05

            for _, s in self.stages.iloc[1:-1].iterrows():
                xtrans = ax.get_xaxis_transform()
                ax.axvline(
                    s.start_seconds / 3600.0,
                    linestyle="dotted",
                    color="black",
                    linewidth=0.5,
                )
                durfrac = (s.end_seconds - s.start_seconds) / totseconds
                if annotate_stage_lines and (durfrac > annotate_frac):
                    ax.text(
                        s.start_seconds / 3600.0 + 0.02,
                        0.9,
                        f"stage {s.stage}",
                        transform=xtrans,
                        rotation=90,
                        verticalalignment="top",
                        horizontalalignment="left",
                    )


def _normalize_filters(
    filters: str | FilterSet | Collection[str | FilterSet],
) -> list[FilterSet]:
    if isinstance(filters, str) or isinstance(filters, FilterSet):
        filters = [filters]
    return [FilterSet.fromstring(filter) for filter in filters]


def _gen_label(
    sample: str,
    well: str,
    filter: FilterSet,
    samples: Sequence[str],
    wells: Sequence[str],
    filters: Sequence[FilterSet],
) -> str:
    label = ""
    if len(samples) > 1:
        label = sample
    if len(wells) > 1:
        if len(label) > 0:
            label += f" ({well})"
        else:
            label = str(well)
    if len(filters) > 1:
        if len(label) > 0:
            label += f", {filter}"
        else:
            label = str(filter)

    return label


def _gen_axtitle(
    expname: str,
    stages: Sequence[int] | slice | int,
    samples: Sequence[str],
    wells: Sequence[str],
    filters: Sequence[FilterSet],
) -> str:
    elems: list[str] = []
    if len(samples) == 1:
        elems += samples
    if len(filters) == 1:
        elems += [str(f) for f in filters]

    val = expname
    if stages != slice(None):
        val += f", stage {_pp_seqsliceint(stages)}"

    if len(elems) > 0:
        val += ": " + ", ".join(elems)

    return val



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import pytest
from click.testing import CliRunner

from qslib import Experiment, Machine
from qslib.cli import cli


@pytest.fixture(scope="module")
def exp():
    return Experiment.from_file("tests/test.eds")


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


def test_info(exp, runner: CliRunner):
    result = runner.invoke(
        cli,
        ["info", "tests/test.eds"],
    )
    assert exp.info().rstrip() == result.output.rstrip()


def test_html(exp, tmp_path_factory: pytest.TempPathFactory, runner: CliRunner):
    tp = tmp_path_factory.mktemp("temp_html")
    result = runner.invoke(
        cli,
        ["info-html", "-o", str(tp / "test.html"), "--no-open", "tests/test.eds"],
    )
    assert exp.info_html()[0:100] == open(tp / "test.html").read()[0:100]


# def test_real_setup(runner: CliRunner):
#     result = runner.invoke(
#         cli,
#         [
#             "setup-machine",
#             "-c",
#             "pass_controller",
#             "-a",
#             "pass_admin",
#             "-d",
#             "localhost",
#             "local_controller_password",
#         ],
#     )

#     assert result.return_value == 0

#     m = Machine("localhost")



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

from qslib.scpi_commands import AccessLevel, ArgList, SCPICommand


def test_unknown():
    s = SCPICommand("UNKNown", "arg1", ["hi", 1, 2], "thing that\n needs quoting")

    assert s.specialize() == s

    assert s.to_scpicommand() == s

    assert SCPICommand.from_scpicommand(s) == s

    print(s.to_string())

    assert s == SCPICommand.from_string(s.to_string())


def test_unknown_type():
    with pytest.raises(TypeError):
        s = SCPICommand("TEST", AccessLevel).to_string()  # type: ignore


def test_comment():
    assert (
        SCPICommand.from_string("EXPOSURE m4,x1,quant,500 # test comment\n").comment
        == "test comment"
    )

    assert SCPICommand.from_string("COMMAND '#' b").comment == None


def test_arglist():
    assert ArgList.from_string("-v1=2.0 t1 t2") == ArgList({"v1": 2.0}, ["t1", "t2"])


def test_neq_other():
    assert SCPICommand("Exp") != "Exp"


def test_auto():
    com = 'TESTCOM arg -an="opt with a string"'
    assert SCPICommand(com) == SCPICommand.from_string(com)

    with pytest.raises(ValueError):
        SCPICommand(com, "extra arg")



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

import qslib.qs_is_protocol as qsp
import qslib.qsconnection_async as qsa
from qslib.data import FilterSet


def test_filterset_strings() -> None:
    a = FilterSet(1, 4)

    assert a == FilterSet.fromstring("x1-m4")
    assert a == FilterSet.fromstring("M4_X1")

    assert a.lowerform == "x1-m4"
    assert a.upperform == "M4_X1"

    assert str(a) == a.lowerform


def test_parse_argstring() -> None:
    # pylint: disable=protected-access
    # FIXME: do quotes allow spaces?
    assert qsa._parse_argstring('-arg1=val1 -arg2="val-2"') == {
        "arg1": "val1",
        "arg2": '"val-2"',
    }


def test_validate_command_format() -> None:
    # pylint: disable=protected-access
    with pytest.raises(ValueError):
        qsp._validate_command_format(b"\nText\n<a>")

    with pytest.raises(ValueError):
        qsp._validate_command_format(b"\nText")

    with pytest.raises(ValueError):
        qsp._validate_command_format(b"<a>\n\n")

    with pytest.raises(ValueError):
        qsp._validate_command_format(b"<a>\nText\n</b><a>")

    with pytest.raises(ValueError):
        qsp._validate_command_format(b"<a>\nText\n</a><a>")

    with pytest.raises(ValueError):
        qsp._validate_command_format(b"<a>\nText\n<a></a>")

    assert qsp._validate_command_format(b"Text") is None

    assert qsp._validate_command_format(b"Text\n") is None

    assert (
        qsp._validate_command_format(b"<a>Text<a>Text</a>\n" b"<b>Text\n\n\n</b></a>")
        is None
    )



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

import argparse
import asyncio
import logging
import sys
from pathlib import Path
from typing import List

import toml
from dacite.core import from_dict

from .monitor import Collector, Config
from .version import __version__

__author__ = "Constantine Evans"
__copyright__ = "Constantine Evans"
__license__ = "BSD-3-Clause"

_logger = logging.getLogger(__name__)


# ---- CLI ----
# The functions defined in this section are wrappers around the main Python
# API allowing them to be called directly from the terminal as a CLI
# executable/script.


def parse_args(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="QuantStudio qPCR data monitor for InfluxDB"
    )
    parser.add_argument(
        "--version",
        action="version",
        version="qs-monitor {ver}".format(ver=__version__),
    )
    parser.add_argument(
        dest="config", help="configuration file", type=Path, metavar="CONFIG_TOML"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """

    logformat = "[%(asctime)s] %(levelname)s:%(name)s: %(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%dT%H:%M:%S"
    )


def main(args: List[str]):
    """Wrapper allowing :func:`fib` to be called with string arguments in a CLI fashion

    Instead of returning the value from :func:`fib`, it prints the result to the
    ``stdout`` in a nicely formatted message.

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--verbose", "42"]``).
    """
    parsed_args = parse_args(args)
    setup_logging(parsed_args.loglevel)

    config = dict(toml.load(parsed_args.config))

    collector = Collector(from_dict(Config, config))

    asyncio.run(collector.reliable_monitor())


def run():
    """Calls :func:`main` passing the CLI arguments extracted from :obj:`sys.argv`

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    # ^  This is a guard statement that will prevent the following code from
    #    being executed in the case someone imports this file instead of
    #    executing it as a script.
    #    https://docs.python.org/3/library/__main__.html

    # After installing your project with pip, users can also run your Python
    # modules as scripts via the ``-m`` flag, as defined in PEP 338::
    #
    #     python -m qpcr_data_updater.skeleton 42
    #
    run()



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

from typing import Any, TypeVar

import pytest

from qslib import Experiment, Machine

# def test_drawer(monkeypatch):
#     def patchrun(self, value: str) -> str:
#         if value == "CLOSE":
#             return ""
#         else:
#             raise ValueError

#     monkeypatch.setattr(Machine, "run_command", patchrun)

#     m = Machine("none")

#     m.drawer_close()


# @pytest.mark.parametrize(
#     "function",
#     [
#         ("drawer_position", ()),
#         ("machine_status", ()),
#         ("run_status", ()),
#         ("get_running_protocol", ()),
#         ("list_runs_in_storage", ()),
#     ],
# )
# def test_mach_not_connected(function: tuple[str, tuple[Any]]) -> None:
#     m = Machine("none")
#     with pytest.raises(ConnectionError):
#         getattr(m, function[0]).__call__(*function[1])


# @pytest.mark.parametrize(
#     "function",
#     [
#         ("from_running", (Machine("none"),)),
#         (
#             "from_machine_storage",
#             (
#                 Machine("none"),
#                 "filename",
#             ),
#         ),
#     ],
# )
# def test_expload_not_connected(function: str) -> None:
#     with pytest.raises(ConnectionError):
#         getattr(Experiment, function[0]).__call__(*function[1])



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

"""(Non-machine/protocol) utility functions for other modules."""

from __future__ import annotations

import re
import xml.etree.ElementTree as ElTr
from datetime import datetime
from typing import Sequence


def _find_or_create(
    element: ElTr.Element | ElTr.ElementTree, path: str
) -> ElTr.Element:
    """Find the element at path, or create it."""
    if isinstance(element, ElTr.ElementTree):
        element = element.getroot()
    if (maybe_element := element.find(path)) is not None:
        return maybe_element
    for elemname in path.split("/"):
        if maybe_element := element.find(elemname):
            element = maybe_element
        else:
            element = ElTr.SubElement(element, elemname)
    return element


def _set_or_create(
    element: ElTr.Element | ElTr.ElementTree,
    path: str,
    text: str | None = None,
    **kwargs: str,
) -> ElTr.Element:
    """Find or create, then set, the element at path."""
    element_at_path = _find_or_create(element, path)
    for key, val in kwargs.items():
        element_at_path.attrib[key] = val
    if text is not None:
        element_at_path.text = text
    return element_at_path


def _unwrap_tags(val: str) -> str:
    """Remove outer <quote>content</quote> tags from string."""
    return re.sub(r"^<[^>]+?>\n?(.*)\n?</[^>]+?>$", r"\1", val)


def _nowuuid() -> str:
    """Consistent and legible UUID-like string from data."""
    return datetime.now().strftime("%Y-%m-%dT%H-%M-%S")


def _pp_seqsliceint(sss: Sequence[int] | slice | int) -> str:
    """Reasonable natural English description slice/int/sequence."""
    if isinstance(sss, int):
        return str(sss)
    if isinstance(sss, Sequence):
        return str(list(sss))
    if isinstance(sss, slice):
        if sss.start is None:
            ret_str = f"up to {sss.stop}"
        elif sss.stop is None:
            ret_str = f"{sss.start} onwards"
        else:
            ret_str = f"{sss.start} to {sss.stop}"
        if sss.step is not None:
            ret_str += f" by step {sss.step}"
        return ret_str
    raise TypeError


__all__ = (
    "_find_or_create",
    "_set_or_create",
    "_unwrap_tags",
    "_nowuuid",
    "_pp_seqsliceint",
)



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import asyncio
import uuid
from asyncio.futures import Future

import pytest
import pytest_asyncio

from qslib import *
from qslib.experiment import AlreadyExistsError, MachineBusyError
from qslib.qs_is_protocol import (
    AccessLevelExceeded,
    AuthError,
    InsufficientAccess,
    InvocationError,
)


@pytest.mark.asyncio
async def test_real_insufficientaccess():
    m = Machine("localhost")
    with m:
        with pytest.raises(InsufficientAccess):
            m.run_command("MACRO USER?")


@pytest.mark.asyncio
async def test_real_accesslevelexceeded():
    m = Machine("localhost", max_access_level="Full")
    with m:
        with pytest.raises(AccessLevelExceeded):
            m.set_access_level(AccessLevel.Full)


@pytest.mark.asyncio
async def test_real_autherror():
    m = Machine("localhost", password="aninvalidpassword")
    with pytest.raises(AuthError):
        with m:
            m.run_command("HELP?")


@pytest.mark.asyncio
def test_real_invocationerror():
    m = Machine("localhost")
    with m:
        with pytest.raises(InvocationError):
            m.run_command("HELP? A B")


# @pytest.mark.asyncio
# async def test_real_fakeexperiment():
#     exp = Experiment.from_file("tests/test.eds")
#     # We need better sample arrangements:
#     exp.sample_wells["Sample 1"] = ["A7", "A8"]
#     exp.sample_wells["Sample 2"] = ["A9", "A10"]
#     exp.sample_wells["othersample"] = ["B7"]

#     with Machine("localhost") as m:
#         with m.at_access("Controller"):
#             exp._populate_folder(m)


@pytest.mark.asyncio
async def test_real_experiment():

    from qslib.monitor import Collector, Config

    mon = Collector(Config())

    confut: Future[bool] = asyncio.Future()
    task = mon.monitor(confut)

    ttask = asyncio.tasks.create_task(task)

    await confut

    proto = Protocol(
        [
            Stage.stepped_ramp(
                50, [30, 31, 32, 33, 34, 35], 120, n_steps=5, collect=True
            )
        ],
        filters=["x1-m4", "x3-m5"],
    )

    exp = Experiment(uuid.uuid1().hex, proto, PlateSetup({"s": "A1"}))

    m = Machine("localhost", port=7000, max_access_level="Controller")

    exp.run("localhost", require_drawer_check=False)

    with pytest.raises(
        MachineBusyError, match=r"Machine localhost:7000 is currently busy: .*"
    ):
        exp.run("localhost")

    exp.sync_from_machine(m)

    exp.pause_now()

    await asyncio.sleep(3)

    proto2 = Protocol(
        [
            Stage.stepped_ramp(
                50, [30, 31, 32, 33, 34, 35], 120, n_steps=5, collect=True
            ),
            Stage.stepped_ramp(30, 50, 120, n_steps=5, collect=True),
        ],
        filters=["x1-m4", "x3-m5"],
    )

    assert proto != proto2
    assert proto != 5

    exp.change_protocol(proto2)

    exp2 = Experiment.from_running("localhost")

    # assert exp.protocol == exp2.protocol
    assert exp.name == exp2.name

    exp.resume()

    exp.get_status()

    exp.sync_from_machine(m)

    with m:
        with m.at_access("Controller"):
            m.run_command(f'TextWAit -timeout=7200 Run Ended \\"{exp.name}\\"')

    await asyncio.sleep(2)

    exp.sync_from_machine(m)

    exp3 = Experiment.from_machine("localhost", exp.name)

    assert exp2.name == exp3.name

    rs = m.run_status()

    assert rs.name == "-"

    with m:
        with m.at_access("Controller"):
            await m.connection.compile_eds(exp.name)

    # with pytest.raises(AlreadyExistsError, match=f"Run {exp.name} exists.*"):
    #     exp.runstate = "INIT"
    #     exp.run("localhost")



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from . import protocol
from .experiment import Experiment
from .machine import Machine, MachineStatus, RunStatus
from .plate_setup import PlateSetup, Sample
from .processors import NormRaw, NormToMaxPerWell, NormToMeanPerWell, SmoothEMWMean
from .protocol import CustomStep, Protocol, Stage, Step
from .scpi_commands import AccessLevel
from .version import __version__

__all__ = (
    "AccessLevel",
    "Machine",
    "Experiment",
    "protocol",
    "PlateSetup",
    "Sample",
    "Protocol",
    "Stage",
    "Step",
    "RunStatus",
    "MachineStatus",
    "NormToMeanPerWell",
    "NormToMaxPerWell",
    "NormRaw",
    "SmoothEMWMean",
    "CustomStep",
)
